#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, TYPE_CHECKING

from piquasso._math.validations import all_in_interval
from piquasso._math.linalg import is_selfadjoint, is_skew_symmetric
from piquasso._math.transformations import xpxp_to_xxpp_indices, xxpp_to_xpxp_indices
from piquasso._math.combinatorics import sort_and_get_parity

from piquasso.api.exceptions import InvalidState, PiquassoException
from piquasso.api.state import State
from piquasso.api.config import Config
from piquasso.api.connector import BaseConnector

from .._utils import (
    validate_fermionic_gaussian_hamiltonian,
    get_fermionic_hamiltonian,
    get_omega,
    tensor_product,
)

if TYPE_CHECKING:
    import numpy as np


class GaussianState(State):
    r"""A fermionic Gaussian state."""

    def __init__(
        self,
        d: int,
        connector: BaseConnector,
        config: Optional[Config] = None,
    ) -> None:
        super().__init__(connector=connector, config=config)

        self._d = d

        # The upper left part of the correlation matrix, i.e., `\Gamma^{a^\dagger a}`.
        self._D = self._np.empty(shape=(d, d), dtype=self._config.complex_dtype)
        # The upper right part of the correlation matrix, i.e.,
        # `\Gamma^{a^\dagger a^\dagger}`.
        self._E = self._np.empty(shape=(d, d), dtype=self._config.complex_dtype)

    @property
    def d(self):
        return self._d

    @property
    def covariance_matrix(self):
        r"""The covariance matrix.
        A fermionic Gaussian state can be fully characterized by its covariance matrix
        defined by

        .. math::

            \Sigma_{ij} := -i \operatorname{Tr} (\rho [\mathbf{m}_i, \mathbf{m}_j]) / 2.

        The covariance matrix is a real-valued, skew-symmetric matrix.
        """

        d = self.d
        np = self._connector.np

        ident = np.identity(d)

        D = self._D
        E = self._E

        two_D_plus_E = 2 * (D + E)
        two_D_minus_E = 2 * (D - E)

        indices = xxpp_to_xpxp_indices(d)

        return np.block(
            [
                [two_D_plus_E.imag, -two_D_minus_E.real + ident],
                [two_D_plus_E.real - ident, two_D_minus_E.imag],
            ]
        )[np.ix_(indices, indices)]

    @covariance_matrix.setter
    def covariance_matrix(self, input_covariance_matrix_xpxp):
        d = self.d

        indices = xpxp_to_xxpp_indices(d)

        input_covariance_matrix_xxpp = input_covariance_matrix_xpxp[
            self._connector.fallback_np.ix_(indices, indices)
        ]

        np = self._connector.np

        ident = np.identity(d)

        two_D_plus_E_real = input_covariance_matrix_xxpp[d:, :d] + ident
        two_D_plus_E_imag = input_covariance_matrix_xxpp[:d, :d]

        two_D_minus_E_real = -input_covariance_matrix_xxpp[:d, d:] + ident
        two_D_minus_E_imag = input_covariance_matrix_xxpp[d:, d:]

        D_plus_E_over_2 = (two_D_plus_E_real + 1j * two_D_plus_E_imag) / 4
        D_minus_E_over_2 = (two_D_minus_E_real + 1j * two_D_minus_E_imag) / 4

        self._D = D_plus_E_over_2 + D_minus_E_over_2
        self._E = D_plus_E_over_2 - D_minus_E_over_2

        self.validate()

    @property
    def correlation_matrix(self):
        r"""The correlation matrix in the Dirac basis.

        The correlation matrix is defined by

        .. math::
            \Gamma := \begin{bmatrix}
                \Gamma^{f^\dagger f} & \Gamma^{f^\dagger f^\dagger} \\
                \Gamma^{f f} & \Gamma^{f f^\dagger}
            \end{bmatrix},

        where :math:`\Gamma_{i,j}^{f^\dagger f} = \langle f_i^\dagger f_j \rangle` and
        :math:`\Gamma_{i,j}^{f f} = \langle f_i f_j \rangle`.

        By CAR, we know that
        :math:`\Gamma_{i,j}^{f^\dagger f} = - \overline{\Gamma_{i,j}^{f f^\dagger }}`
        and
        :math:`\Gamma_{i,j}^{f f} = I - \overline{\Gamma_{i,j}^{f^\dagger f^\dagger }}`

        The correlation matrix is a self-adjoint matrix.
        """

        np = self._connector.np

        ident = np.identity(self.d)

        return np.block([[self._D, self._E], [-self._E.conj(), ident - self._D.conj()]])

    @correlation_matrix.setter
    def correlation_matrix(self, input_correlation_matrix):
        d = self.d
        self._D = input_correlation_matrix[:d, :d]
        self._E = input_correlation_matrix[:d, d:]
        self.validate()

    @property
    def maj_correlation_matrix(self):
        """The correlation matrix in the Majorana operator basis."""
        d = self.d
        np = self._connector.np

        ident = np.identity(d)

        D = self._D
        E = self._E

        two_D_plus_E = 2 * (D + E)
        two_D_minus_E = 2 * (D - E)

        return (
            np.block(
                [
                    [
                        ident + 1j * np.imag(two_D_plus_E),
                        1j * (ident - np.real(two_D_minus_E)),
                    ],
                    [
                        1j * (np.real(two_D_plus_E) - ident),
                        ident + 1j * np.imag(two_D_minus_E),
                    ],
                ]
            )
        ) / 2

    def validate(self):
        if not self._config.validate:
            return

        d = self.d

        np = self._connector.fallback_np

        correlation_matrix = self.correlation_matrix

        expected_shape = (2 * d, 2 * d)
        actual_shape = correlation_matrix.shape

        if not actual_shape == expected_shape:
            raise InvalidState(
                f"The correlation matrix has invalid shape.\n"
                f"expected={expected_shape} "
                f"actual={actual_shape}"
            )

        if not np.allclose(
            correlation_matrix[:d, :d],
            np.identity(d) - correlation_matrix[d:, d:].conj(),
        ) or not np.allclose(
            correlation_matrix[:d, d:], -correlation_matrix[d:, :d].conj()
        ):
            raise InvalidState("The correlation matrix is not physical.")

        if not is_selfadjoint(correlation_matrix[:d, :d]) or not is_skew_symmetric(
            correlation_matrix[:d, d:]
        ):
            raise InvalidState("The correlation matrix is not self-adjoint.")

        eigvals = np.linalg.eigvalsh(correlation_matrix)

        if not all_in_interval(eigvals, 0.0, 1.0):
            raise InvalidState("The correlation matrix has invalid eigenvalues.")

    @property
    def fock_probabilities(self) -> "np.ndarray":
        d = self.d
        probabilities = []

        fallback_np = self._connector.fallback_np

        for i in range(2**d):
            occupation_numbers = fallback_np.empty(d, dtype=int)

            for j in range(d):
                occupation_numbers[d - j - 1] = i % 2
                i = i // 2

            probabilities.append(
                self.get_particle_detection_probability(occupation_numbers)
            )

        return self._connector.np.array(probabilities)

    def get_particle_detection_probability(
        self, occupation_number: "np.ndarray"
    ) -> float:
        basis_state = self.__class__(
            d=self._d, connector=self._connector, config=self._config
        )
        basis_state._set_occupation_numbers(occupation_number)

        return self.overlap(basis_state)

    def mean_particle_numbers(self, modes):
        """Returns the mean particle numbers on the specified modes."""
        np = self._connector.np

        return np.real(np.diag(self._D)[modes,])

    def _set_parent_hamiltonian(self, parent_hamiltonian):
        """Constructs the Gaussian state using its parent Hamiltonian."""

        if self._config.validate:
            validate_fermionic_gaussian_hamiltonian(parent_hamiltonian)

        d = len(parent_hamiltonian) // 2

        np = self._connector.np

        self.correlation_matrix = np.linalg.inv(
            np.identity(2 * d) + self._connector.expm(2 * parent_hamiltonian)
        )

    def get_parent_hamiltonian(self):
        r"""Calculates the parent Hamiltonian.

        When the correlation matrix is not singular, the density matrix is

        .. math::
            \rho = e^{\hat{H}} / \operatorname{Tr} e^{\hat{H}},

        where :math:`\hat{H}` is the parent hamiltonian (see
        :meth:`get_parent_hamiltonian`) given by

        .. math::
            \hat{H} = \mathbf{f}^\dagger H \mathbf{f}, \\\\

            H = \begin{bmatrix}
                A & -\overline{B} \\
                B & -\overline{A}
            \end{bmatrix}.

        where :math:`A^\dagger = A` and :math:`B^T = - B`.

        Raises:
            PiquassoException: When the correlation matrix is singular.
        """
        d = self.d
        np = self._connector.np

        if (
            self._config.validate
            and np.linalg.matrix_rank(self.correlation_matrix) != 2 * self.d
        ):
            raise PiquassoException(
                "Cannot calculate parent Hamiltonian, since the correlation matrix is "
                "singular."
            )

        inverse_correlation_matrix = np.linalg.inv(self.correlation_matrix)

        parent_hamiltonian = (
            self._connector.logm(inverse_correlation_matrix - np.identity(2 * d)) / 2
        )

        return parent_hamiltonian

    @property
    def density_matrix(self) -> "np.ndarray":
        r"""Density matrix in the lexicographic ordering.

        When applicable, the density matrix is just

        .. math::
            \rho = e^{\hat{H}} / \operatorname{Tr} e^{\hat{H}},

        where :math:`\hat{H}` is the parent hamiltonian (see
        :meth:`get_parent_hamiltonian`) given by

        .. math::
            \hat{H} = \mathbf{f}^\dagger H \mathbf{f}, \\\\
            H = \begin{bmatrix}
                A & -\overline{B} \\
                B & -\overline{A}
            \end{bmatrix}.

        where :math:`A^\dagger = A` and :math:`B^T = - B`.

        Note:
            The density matrix is returned in matrix form here. This will probably
            change in the future to tensor form.
        """

        d = self.d
        connector = self._connector
        np = self._connector.np

        blocks, O = self._connector.schur(self.covariance_matrix)

        etas = np.diag(blocks, 1)[::2].copy()

        is_special_orthogonal: bool = np.isclose(np.linalg.det(O), 1.0)

        if not is_special_orthogonal:
            # NOTE: If the orthogonal coming from the real Schur decomposition is not
            # special, then there is no corresponding real logarithm from which we can
            # calculate the gate hamiltonian. To fix this, we can just perform a flip.
            flip = connector.block_diag(
                *([np.array([[0, 1], [1, 0]])] + [np.identity(2)] * (d - 1))
            )
            O = O @ flip
            etas = connector.assign(etas, 0, -etas[0])

        # 1. Calculate diagonalized density matrix
        nu = (1 + etas) / 2

        single_mode_dms = []

        for i in range(self.d):
            single_mode_dms.append(np.array([[nu[i], 0], [0, 1 - nu[i]]]))

        rho_D = tensor_product(single_mode_dms)

        # 2. Unitary which diagonalizes the original density matrix
        omega = get_omega(self.d)

        indices = xxpp_to_xpxp_indices(d)

        basis_transformation = omega[indices]

        i2H = (
            basis_transformation.conj().T
            @ connector.real_logm(O)
            @ basis_transformation
        )

        H = i2H / 2j
        bigH = get_fermionic_hamiltonian(H, connector)
        unitary = connector.expm(1j * bigH)

        return unitary @ rho_D @ unitary.conj().T

    def _set_occupation_numbers(self, occupation_numbers):
        dtype = self._config.dtype
        fallback_np = self._connector.fallback_np
        np = self._connector.np

        plus_minus = np.diag(1 - 2 * np.array(occupation_numbers))

        zeros = np.zeros_like(plus_minus)

        d = len(occupation_numbers)

        indices = xxpp_to_xpxp_indices(d)

        self.covariance_matrix = (
            np.block([[zeros, plus_minus], [-plus_minus, zeros]])[
                fallback_np.ix_(indices, indices)
            ]
        ).astype(dtype)

    @classmethod
    def _from_representation(
        cls,
        *,
        D: "np.ndarray",
        E: "np.ndarray",
        config: Config,
        connector: BaseConnector,
    ) -> "GaussianState":
        obj = cls(d=len(D), connector=connector, config=config)

        obj._D = D
        obj._E = E

        return obj

    def reduced(self, modes):
        """Reduces the state to a subsystem on the specified modes."""

        index = self._connector.fallback_np.ix_(modes, modes)

        return self._from_representation(
            D=self._D[index],
            E=self._E[index],
            config=self._config,
            connector=self._connector,
        )

    def get_majorana_monomial_expectation_value(self, indices):
        r"""Calculates Majorana monomial expectation values using Wick's theorem.

        The monomial indices are understood in the xxpp-ordering.

        Args:
            indices:
                Indices of the Majorana operators in any order with possible with
                possible multiplicities.
        """

        fallback_np = self._connector.fallback_np

        indices = fallback_np.array(indices)

        sorted_indices, parity = sort_and_get_parity(indices)

        unique, counts = fallback_np.unique(sorted_indices, return_counts=True)
        filtered_indices = unique[counts % 2 == 1]

        matrix_index = fallback_np.ix_(filtered_indices, filtered_indices)

        covariance_matrix_reduced = self.covariance_matrix[matrix_index]

        prefactor = parity * 1j ** (len(filtered_indices) // 2)

        return prefactor * self._connector.pfaffian(covariance_matrix_reduced)

    def get_parity_operator_expectation_value(self):
        r"""Calculates the expectation value of the parity operator.

        The parity operator is defined as

        .. math::
            P = i^d m_1 \dots m_{2d} = i^d x_1 \dots x_d p_1 \dots p_d.
        """
        fallback_np = self._connector.fallback_np

        return (-1j) ** self.d * self.get_majorana_monomial_expectation_value(
            fallback_np.arange(2 * self.d)
        )

    def overlap(self, other):
        r"""Calculates the overlap between two fermionic Gaussian states.

        The overlap is the Hilbert-Schmidt inner product of the density matrices.
        """

        np = self._connector.np

        gamma1 = self.covariance_matrix
        gamma2 = other.covariance_matrix

        ident = np.identity(len(gamma1))

        ret2 = np.sqrt(np.linalg.det((ident - gamma1 @ gamma2) / 2))

        return ret2

    def __eq__(self, other):
        np = self._connector.np
        return (
            type(self) is type(other)
            and np.allclose(self._D, other._D)
            and np.allclose(self._E, other._E)
        )



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import numpy as np

import piquasso as pq

from scipy.linalg import expm

from piquasso.decompositions.clements import clements, instructions_from_decomposition


for_all_connectors = pytest.mark.parametrize(
    "connector", [pq.NumpyConnector(), pq.JaxConnector()]
)


@pytest.mark.monkey
@for_all_connectors
def test_GaussianHamiltonian_and_Interferometer_equivalence(
    connector,
    generate_hermitian_matrix,
):
    d = 3

    A = generate_hermitian_matrix(d)
    zeros = np.zeros_like(A)

    unitary = expm(-2j * A.conj())

    hamiltonian = np.block([[A, zeros], [zeros, -A.conj()]])

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state_vector = [1, 0, 1]

    with pq.Program() as program:
        pq.Q() | pq.StateVector(state_vector)

        pq.Q() | pq.fermionic.GaussianHamiltonian(hamiltonian=hamiltonian)

    with pq.Program() as passive_program:
        pq.Q() | pq.StateVector(state_vector)

        pq.Q() | pq.Interferometer(matrix=unitary)

    state_1 = simulator.execute(program).state
    state_2 = simulator.execute(passive_program).state

    assert state_1 == state_2


@pytest.mark.monkey
@for_all_connectors
def test_passive_GaussianHamiltonian_preserves_particle_number(
    connector,
    generate_passive_fermionic_gaussian_hamiltonian,
):
    d = 3

    hamiltonian = generate_passive_fermionic_gaussian_hamiltonian(d)

    state_vector = [1, 0, 1]

    with pq.Program() as program:
        pq.Q() | pq.StateVector(state_vector)

        pq.Q() | pq.fermionic.GaussianHamiltonian(hamiltonian=hamiltonian)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state = simulator.execute(program).state

    mean_particle_numberss = state.mean_particle_numbers(modes=(0, 1, 2))

    assert (
        mean_particle_numberss.dtype == simulator.config.dtype
    ), "Particle number should be real-valued"
    assert np.isclose(
        sum(mean_particle_numberss),
        sum(state_vector),
    )


@pytest.mark.monkey
@for_all_connectors
def test_GaussianHamiltonian_subsystem_equivalence(
    connector,
    generate_fermionic_gaussian_hamiltonian,
):
    d = 3

    modes = (0, 2)

    hamiltonian = generate_fermionic_gaussian_hamiltonian(d - 1)

    embedded_hamiltonian = connector.np.zeros((2 * d, 2 * d), dtype=hamiltonian.dtype)

    doubled_modes = np.concatenate([modes, np.array(modes) + d])

    embedded_hamiltonian = connector.assign(
        embedded_hamiltonian, np.ix_(doubled_modes, doubled_modes), hamiltonian
    )

    state_vector = [1, 0, 1]
    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    with pq.Program() as program_subsystem:
        pq.Q() | pq.StateVector(state_vector)

        pq.Q(*modes) | pq.fermionic.GaussianHamiltonian(hamiltonian=hamiltonian)

    with pq.Program() as program_embedded:
        pq.Q() | pq.StateVector(state_vector)

        pq.Q() | pq.fermionic.GaussianHamiltonian(hamiltonian=embedded_hamiltonian)

    subsystem_state = simulator.execute(program_subsystem).state
    embedded_state = simulator.execute(program_embedded).state

    assert subsystem_state == embedded_state


@for_all_connectors
def test_Interferometer_on_state_vector(connector):
    passive_hamiltonian = np.array([[1, 2j, 3j], [-2j, 5, 6], [-3j, 6, 7]])

    U = expm(1j * passive_hamiltonian)

    state_vector = [1, 0, 1]

    with pq.Program() as program:
        pq.Q() | pq.StateVector(state_vector)

        pq.Q() | pq.Interferometer(U)

    simulator = pq.fermionic.GaussianSimulator(d=3, connector=connector)

    state = simulator.execute(program).state

    expected_covariance_matrix = np.array(
        [
            [0.0, -0.932682, -0.156675, 0.30532, -0.106053, -0.033002],
            [0.932682, 0.0, -0.30532, -0.156675, 0.033002, -0.106053],
            [0.156675, 0.30532, 0.0, 0.749426, -0.557812, 0.097148],
            [-0.30532, 0.156675, -0.749426, 0.0, -0.097148, -0.557812],
            [0.106053, -0.033002, 0.557812, 0.097148, 0.0, -0.816744],
            [0.033002, 0.106053, -0.097148, 0.557812, 0.816744, 0.0],
        ]
    )

    assert np.allclose(
        state.covariance_matrix,
        expected_covariance_matrix,
    )


@for_all_connectors
def test_Interferometer(connector):
    d = 3

    A = np.array([[1, 2j, 3j], [-2j, 5, 6], [-3j, 6, 7]])
    B = np.array([[0, 1j, 2], [-1j, 0, 3], [-2, -3, 0]])

    hamiltonian = np.block([[-A.conj(), B], [-B.conj(), A]])

    passive_hamiltonian = np.array([[1, 2j, 3j], [-2j, 5, 6], [-3j, 6, 7]])

    U = expm(1j * passive_hamiltonian)

    state_vector = [1, 0, 1]

    with pq.Program() as program:
        pq.Q() | pq.StateVector(occupation_numbers=state_vector)

        pq.Q() | pq.fermionic.GaussianHamiltonian(hamiltonian=hamiltonian)

        pq.Q() | pq.Interferometer(matrix=U)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state = simulator.execute(program).state

    expected_covariance_matrix = np.array(
        [
            [0.0, 0.469966, -0.19859, 0.780816, 0.228407, 0.279019],
            [-0.469966, 0.0, -0.66337, 0.08203, -0.493825, -0.297456],
            [0.19859, 0.66337, 0.0, -0.392254, -0.478024, 0.371661],
            [-0.780816, -0.08203, 0.392254, 0.0, -0.069497, 0.474242],
            [-0.228407, 0.493825, 0.478024, 0.069497, 0.0, -0.686026],
            [-0.279019, 0.297456, -0.371661, -0.474242, 0.686026, 0.0],
        ]
    )

    assert np.allclose(state.covariance_matrix, expected_covariance_matrix)


@for_all_connectors
def test_Interferometer_on_subsystem(connector):
    d = 3

    A = np.array([[1, 2j, 3j], [-2j, 5, 6], [-3j, 6, 7]])
    B = np.array([[0, 1j, 2], [-1j, 0, 3], [-2, -3, 0]])

    hamiltonian = np.block([[-A.conj(), B], [-B.conj(), A]])

    passive_hamiltonian = np.array([[1, 2j], [-2j, 5]])

    U = expm(1j * passive_hamiltonian)

    state_vector = [1, 0, 1]

    with pq.Program() as program:
        pq.Q() | pq.StateVector(occupation_numbers=state_vector)

        pq.Q() | pq.fermionic.GaussianHamiltonian(hamiltonian=hamiltonian)

        pq.Q(0, 2) | pq.Interferometer(matrix=U)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state = simulator.execute(program).state

    expected_covariance_matrix = np.array(
        [
            [0.0, 0.16447, 0.252142, 0.588858, 0.365978, -0.654737],
            [-0.16447, 0.0, -0.951866, 0.162249, 0.145479, -0.139325],
            [-0.252142, 0.951866, -0.0, -0.147157, -0.040228, 0.084272],
            [-0.588858, -0.162249, 0.147157, -0.0, 0.672212, 0.391659],
            [-0.365978, -0.145479, 0.040228, -0.672212, 0.0, -0.625626],
            [0.654737, 0.139325, -0.084272, -0.391659, 0.625626, 0.0],
        ]
    )

    assert np.allclose(state.covariance_matrix, expected_covariance_matrix)


@pytest.mark.monkey
@for_all_connectors
def test_Beamsplitter_preserves_particle_number(connector):
    d = 3

    state_vector = [1, 0, 1]

    with pq.Program() as program:
        pq.Q() | pq.StateVector(state_vector)

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 3, phi=np.pi / 4)
        pq.Q(1, 2) | pq.Beamsplitter(theta=-np.pi / 5, phi=np.pi / 4)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state = simulator.execute(program).state

    assert np.isclose(
        sum(state.mean_particle_numbers(modes=(0, 1, 2))),
        sum(state_vector),
    )


@for_all_connectors
def test_5050_Beamsplitter(connector):
    d = 2
    state_vector = [1, 0]

    with pq.Program() as program:
        pq.Q() | pq.StateVector(state_vector)

        pq.Q() | pq.Beamsplitter(theta=np.pi / 4)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state = simulator.execute(program).state

    assert np.allclose(
        state.correlation_matrix,
        np.array(
            [
                [0.5, 0.5, 0.0, 0.0],
                [0.5, 0.5, 0.0, 0.0],
                [0.0, 0.0, 0.5, -0.5],
                [0.0, 0.0, -0.5, 0.5],
            ]
        ),
    )


@for_all_connectors
def test_Interferometer_clements_equivalence(connector):
    d = 3

    A = np.array([[1, 2j, 3j], [-2j, 5, 6], [-3j, 6, 7]])
    B = np.array([[0, 1j, 2], [-1j, 0, 3], [-2, -3, 0]])

    hamiltonian = np.block([[-A.conj(), B], [-B.conj(), A]])

    passive_hamiltonian = np.array([[1, 2j, 3j], [-2j, 5, 6], [-3j, 6, 7]])

    U = expm(1j * passive_hamiltonian)

    with pq.Program() as preparation:
        pq.Q() | pq.StateVector(occupation_numbers=[1, 0, 1])
        pq.Q() | pq.fermionic.GaussianHamiltonian(hamiltonian=hamiltonian)

    with pq.Program() as program:
        pq.Q() | preparation
        pq.Q() | pq.Interferometer(matrix=U)

    decomposition = clements(U, connector=connector)

    with pq.Program() as decomposed_program:
        pq.Q() | preparation
        decomposed_program.instructions.extend(
            instructions_from_decomposition(decomposition)
        )

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state = simulator.execute(program).state
    decomposed_state = simulator.execute(decomposed_program).state

    assert state == decomposed_state


@for_all_connectors
def test_Interferometer_subsystem_equivalence(connector, generate_unitary_matrix):
    d = 3

    U = generate_unitary_matrix(2)

    bigU = np.identity(d, dtype=complex)

    bigU[0, 0] = U[0, 0]
    bigU[0, 2] = U[0, 1]
    bigU[2, 0] = U[1, 0]
    bigU[2, 2] = U[1, 1]

    with pq.Program() as preparation:
        pq.Q() | pq.StateVector(occupation_numbers=[1, 1, 0])

    with pq.Program() as program_subsystem:
        pq.Q() | preparation
        pq.Q(0, 2) | pq.Interferometer(matrix=U)

    with pq.Program() as program_full:
        pq.Q() | preparation
        pq.Q(0, 1, 2) | pq.Interferometer(matrix=bigU)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state_subsystem = simulator.execute(program_subsystem).state
    state_full = simulator.execute(program_full).state

    assert state_subsystem == state_full


@for_all_connectors
def test_Squeezing2_on_two_modes_00(connector):
    d = 2

    state_vector = [0, 0]

    r = 0.1
    phi = np.pi / 7

    with pq.Program() as program:
        pq.Q() | pq.StateVector(occupation_numbers=state_vector)

        pq.Q(0, 1) | pq.Squeezing2(r=r, phi=phi)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state = simulator.execute(program).state

    term_00 = np.cos(r / 2)
    term_11 = -np.sin(r / 2) * np.exp(1j * phi)

    expected_state_vector = np.array([term_00, 0.0, 0.0, term_11])

    expected_density_matrix = np.outer(
        expected_state_vector, expected_state_vector.conj()
    )

    actual_density_matrix = state.density_matrix

    assert np.allclose(expected_density_matrix, actual_density_matrix)


@for_all_connectors
def test_Squeezing2_on_two_modes_11(connector):
    d = 2

    state_vector = [1, 1]

    r = 0.1
    phi = np.pi / 7

    with pq.Program() as program:
        pq.Q() | pq.StateVector(occupation_numbers=state_vector)

        pq.Q(0, 1) | pq.Squeezing2(r=r, phi=phi)

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    state = simulator.execute(program).state

    term_00 = np.sin(r / 2) * np.exp(-1j * phi)
    term_11 = np.cos(r / 2)

    expected_state_vector = np.array([term_00, 0.0, 0.0, term_11])

    expected_density_matrix = np.outer(
        expected_state_vector, expected_state_vector.conj()
    )

    actual_density_matrix = state.density_matrix

    assert np.allclose(expected_density_matrix, actual_density_matrix)


@for_all_connectors
def test_Squeezing2_leaves_odd_occupation_numbers_invariant(connector):
    d = 2

    state_vector = [1, 0]

    r = 0.1
    phi = np.pi / 7

    simulator = pq.fermionic.GaussianSimulator(d=d, connector=connector)

    with pq.Program() as empty_program:
        pq.Q() | pq.StateVector(occupation_numbers=state_vector)

    with pq.Program() as program:
        pq.Q() | empty_program

        pq.Q(0, 1) | pq.Squeezing2(r=r, phi=phi)

    initial_state = simulator.execute(empty_program).state
    squeezed_state = simulator.execute(program).state

    assert initial_state == squeezed_state



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax

import jax.numpy as jnp

from piquasso._math.linalg import reduce_

from .utils import factorial, eig


def loop_hafnian_with_reduction(matrix, diagonal, reduce_on):
    """JAX implementation of the loop hafnian.

    This functions implement the loop hafnian in JAX with Glynn-type iterations, because
    this way the size of the matrix stays constant, which is required by JAX. Also, the
    speedup using repetitions cannot be utilized, since that would require dynamic for
    loops.

    This is just a placeholder implementation. To properly implement this, one would
    need to use JAX FFI.
    """
    reduced_matrix = _reduce_matrix_with_diagonal(matrix, diagonal, reduce_on)

    return _loop_hafnian(reduced_matrix)


def _reduce_matrix_with_diagonal(matrix, diagonal, reduce_on):
    reduced_diagonal = reduce_(diagonal, jnp.array(reduce_on))
    reduced_matrix = reduce_(matrix, jnp.array(reduce_on))

    for i in range(len(reduced_diagonal)):
        reduced_matrix = reduced_matrix.at[i, i].set(reduced_diagonal[i])

    return reduced_matrix


def _calculate_f(polynom_coefficients):
    dim_over_2 = len(polynom_coefficients)
    dim = 2 * dim_over_2

    data = jnp.zeros(shape=(2, dim_over_2 + 1), dtype=polynom_coefficients.dtype)

    data = data.at[0, 0].set(1.0)
    indices = jnp.arange(dim_over_2 + 1)

    def outer_body_fun(idx, args):
        data, p_aux1 = args
        factor = polynom_coefficients[idx - 1]

        p_aux0 = jax.lax.cond(idx % 2 == 1, lambda: 0, lambda: 1)
        p_aux1 = jax.lax.cond(idx % 2 == 1, lambda: 1, lambda: 0)

        data = data.at[p_aux1].set(data[p_aux0])

        size = dim // (2 * idx) + 1

        powfactors = jnp.power(factor, jnp.arange(dim_over_2 + 1)) / factorial(
            jnp.arange(dim_over_2 + 1, dtype=float)
        )

        def inner_body_fun(jdx, data):
            mask = jnp.where(idx * jdx <= indices, 1, 0)
            aux0_copy = jnp.copy(data[p_aux0])
            aux0_rolled = jnp.roll(aux0_copy, idx * jdx)

            mask2 = jnp.where(jdx < size, 1, 0)
            return data * (1 - mask2) + mask2 * data.at[p_aux1].set(
                data[p_aux1] + aux0_rolled * mask * powfactors[jdx]
            )

        data = jax.lax.fori_loop(1, dim_over_2 + 1, inner_body_fun, data)

        return (data, p_aux1)

    data, p_aux1 = jax.lax.fori_loop(1, dim_over_2 + 1, outer_body_fun, (data, 1))

    return data[p_aux1]


@jax.jit
def _loop_hafnian(A):
    if len(A) % 2 == 1:
        # NOTE: If the ijnput matrix `A` has an odd dimension, e.g. 7x7, then the matrix
        # should be padded to an even dimension, to e.g. 8x8.
        A = jnp.pad(A, pad_width=((1, 0), (1, 0)))
        A[0, 0] = 1.0

    degree = A.shape[0] // 2

    if degree == 0:
        return 1.0

    ret = 0.0

    size = 2**degree

    def body_fun(permutation_idx, ret):
        prefact = 1

        delta = jnp.empty(degree, dtype=int)

        def func_delta(i, args):
            j, delta = args
            delta = delta.at[i].set(2 * (j % 2) - 1)

            return j // 2, delta

        delta = jax.lax.fori_loop(0, degree, func_delta, (permutation_idx, delta))[1]

        prefact = jnp.prod(delta)

        O = jnp.zeros(shape=(degree, degree), dtype=A.dtype)
        diag_delta = jnp.diag(delta)
        X_delta = jnp.block([[O, diag_delta], [diag_delta, O]])

        polynom_coefficients = _get_loop_polynom_coefficients(A, X_delta, degree)

        ret += prefact * _calculate_f(polynom_coefficients)[degree]

        return ret

    ret = jax.lax.fori_loop(0, size, body_fun, ret)

    return ret / size


def _get_loop_polynom_coefficients(A, X_delta, degree):
    AX_delta = A @ X_delta

    eigenvalues, O = eig(AX_delta)

    Oinv = jnp.linalg.inv(O)

    v = jnp.diag(A)

    left = v @ X_delta @ O
    right = Oinv @ v.T

    def func(power):
        diags = jnp.diag(jnp.power(eigenvalues, power - 1))
        powertrace = jnp.sum(jnp.power(eigenvalues, power))
        return (powertrace / power + (left @ diags @ right)) / 2.0

    return jax.lax.map(func, jnp.arange(1, degree + 1))



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import piquasso as pq


def test_single_mode_single_instruction_registry(FakeGate):
    with pq.Program() as program:
        pq.Q(0) | FakeGate(dummyparam=420)

    assert len(program.instructions) == 1

    assert program.instructions[0].modes == (0,)
    assert program.instructions[0].params == {"dummyparam": 420}


def test_single_mode_multiple_instruction_registry(FakeGate):
    with pq.Program() as program:
        pq.Q(0, 1) | FakeGate(dummyparam=420) | FakeGate(
            dummyparam1=42, dummyparam2=320
        )

    assert len(program.instructions) == 2

    assert program.instructions[0].modes == (0, 1)
    assert program.instructions[0].params == {"dummyparam": 420}

    assert program.instructions[1].modes == (0, 1)
    assert program.instructions[1].params == {
        "dummyparam1": 42,
        "dummyparam2": 320,
    }


def test_multiple_mode_single_instructionregistry(FakeGate):
    with pq.Program() as program:
        pq.Q(2, 1, 0) | FakeGate(dummyparam1=421)
        pq.Q(1) | FakeGate(dummyparam2=1)
        pq.Q(0, 2) | FakeGate(dummyparam3=999)

    assert len(program.instructions) == 3

    assert program.instructions[0].modes == (2, 1, 0)
    assert program.instructions[0].params == {"dummyparam1": 421}

    assert program.instructions[1].modes == (1,)
    assert program.instructions[1].params == {"dummyparam2": 1}

    assert program.instructions[2].modes == (0, 2)
    assert program.instructions[2].params == {"dummyparam3": 999}


def test_multiple_mode_multiple_instructionregistry(FakeGate):
    with pq.Program() as program:
        pq.Q(4) | FakeGate(param=2) | FakeGate(param=0)
        pq.Q(0, 2) | FakeGate(param=999)
        pq.Q(1, 0) | FakeGate(param=1) | FakeGate(param=9)

    assert len(program.instructions) == 5

    assert program.instructions[0].modes == (4,)
    assert program.instructions[0].params == {"param": 2}

    assert program.instructions[1].modes == (4,)
    assert program.instructions[1].params == {"param": 0}

    assert program.instructions[2].modes == (0, 2)
    assert program.instructions[2].params == {"param": 999}

    assert program.instructions[3].modes == (1, 0)
    assert program.instructions[3].params == {"param": 1}

    assert program.instructions[4].modes == (1, 0)
    assert program.instructions[4].params == {"param": 9}


def test_instruction_registration_with_no_modes_is_resolved_to_all_modes(
    FakeGate,
    FakeSimulator,
):
    with pq.Program() as program:
        pq.Q() | FakeGate(param="some-parameter")

    simulator = FakeSimulator(d=42)
    state = simulator.execute(program).state

    assert program.instructions[0].modes == tuple(range(state.d))


def test_instruction_registration_with_all_keyword_is_resolved_to_all_modes(
    FakeGate,
    FakeSimulator,
):
    with pq.Program() as program:
        pq.Q(all) | FakeGate(param="some-parameter")

    simulator = FakeSimulator(d=42)
    state = simulator.execute(program).state

    assert program.instructions[0].modes == tuple(range(state.d))



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import piquasso as pq


pytestmark = pytest.mark.benchmark(
    group="pure-fock",
)


@pytest.fixture
def r():
    return 0.6


@pytest.fixture
def cutoff():
    return 10


@pytest.mark.parametrize("d", (3, 4, 5, 6))
def piquasso_benchmark(benchmark, d, cutoff, r):
    @benchmark
    def func():
        with pq.Program() as program:
            pq.Q() | pq.Vacuum()

            for i in range(d):
                pq.Q(i) | pq.Squeezing(r=r)

        simulator_fock = pq.PureFockSimulator(d=d, config=pq.Config(cutoff=cutoff))

        simulator_fock.execute(program)


@pytest.mark.parametrize("d", (3, 4, 5, 6))
def strawberryfields_benchmark(benchmark, d, cutoff, r, sf):
    @benchmark
    def func():
        eng = sf.Engine(backend="fock", backend_options={"cutoff_dim": cutoff})

        circuit = sf.Program(d)

        with circuit.context as q:
            for w in range(d):
                sf.ops.Sgate(r) | q[w]

        eng.run(circuit).state



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from piquasso.instructions import gates, preparations

from ..instructions import GaussianHamiltonian, ParentHamiltonian, IsingXX

from .calculations import (
    state_vector,
    vacuum,
    parent_hamiltonian,
    gaussian_hamiltonian,
    passive_linear_gate,
    squeezing2,
    ising_XX,
)
from .state import GaussianState

from piquasso._simulators.simulator import BuiltinSimulator
from piquasso._simulators.connectors import NumpyConnector, JaxConnector


class GaussianSimulator(BuiltinSimulator):
    """
    Fermionic Gaussian simulator.

    Example usage::

        passive_hamiltonian = np.array([[1, 2j, 3j], [-2j, 5, 6], [-3j, 6, 7]])

        U = expm(1j * passive_hamiltonian)

        state_vector = [1, 0, 1]

        with pq.Program() as program:
            pq.Q() | pq.StateVector(state_vector)

            pq.Q() | pq.Interferometer(U)

        simulator = pq.fermionic.GaussianSimulator(d=3, connector=connector)

        state = simulator.execute(program).state

    Supported preparations:
        :class:`~piquasso.instructions.preparations.Vacuum`,
        :class:`~piquasso.instructions.preparations.StateVector`.

    Supported gates:
        :class:`~piquasso.instructions.gates.Interferometer`,
        :class:`~piquasso.instructions.gates.Beamsplitter`,
        :class:`~piquasso.instructions.gates.Phaseshifter`,
        :class:`~piquasso.fermionic.instructions.GaussianHamiltonian`.
    """

    _state_class = GaussianState

    _instruction_map = {
        preparations.Vacuum: vacuum,
        preparations.StateVector: state_vector,
        ParentHamiltonian: parent_hamiltonian,
        GaussianHamiltonian: gaussian_hamiltonian,
        gates.Interferometer: passive_linear_gate,
        gates.Beamsplitter: passive_linear_gate,
        gates.Phaseshifter: passive_linear_gate,
        gates.Squeezing2: squeezing2,
        IsingXX: ising_XX,
    }

    _default_connector_class = NumpyConnector

    _extra_builtin_connectors = [JaxConnector]



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Benchmarking the Piquasso and TheWalrus torontonian implementations.
"""

import time

import piquasso as pq

import json

from piquasso._math.torontonian import torontonian as pq_torontonian
from thewalrus import rec_torontonian as tw_torontonian

import matplotlib.pyplot as plt

from scipy.stats import unitary_group

import numpy as np

np.set_printoptions(suppress=True, linewidth=200)


if __name__ == "__main__":
    x = []
    y = []
    z = []

    ITER = 100

    FILENAME = f"torontonian_benchmark_{int(time.time())}.json"

    for d in range(3, 30):
        print(d)
        x.append(d)

        simulator = pq.GaussianSimulator(d=d)

        program = pq.Program(
            instructions=[pq.Vacuum()]
            + [pq.Squeezing(r=np.random.rand()).on_modes(i) for i in range(d)]
            + [pq.Interferometer(unitary_group.rvs(d))]
        )

        state = simulator.execute(program).state

        xpxp_covariance_matrix = state.xpxp_covariance_matrix

        sigma: np.ndarray = (xpxp_covariance_matrix / 2 + np.identity(2 * d)) / 2

        input_matrix = np.identity(len(sigma), dtype=float) - np.linalg.inv(sigma)

        sum_ = 0.0

        print(pq_torontonian(input_matrix))
        for _ in range(ITER):
            print("|", end="", flush=True)
            start_time = time.time()
            pq_torontonian(input_matrix)
            sum_ += time.time() - start_time

        y.append(sum_ / ITER)
        print()

        xxpp_covariance_matrix = state.xxpp_covariance_matrix

        sigma: np.ndarray = (xxpp_covariance_matrix / 2 + np.identity(2 * d)) / 2

        input_matrix = np.identity(len(sigma), dtype=float) - np.linalg.inv(sigma)

        sum_ = 0.0

        print(tw_torontonian(input_matrix))
        for _ in range(ITER):
            print("-", end="", flush=True)
            start_time = time.time()
            tw_torontonian(input_matrix)
            sum_ += time.time() - start_time

        z.append(sum_ / ITER)
        print()

        with open(FILENAME, "w") as f:
            json.dump(dict(x=x, y=y, z=z), f, indent=4)

    plt.scatter(x, y, label="Piquasso")
    plt.scatter(x, z, label="TheWalrus")
    plt.legend()
    plt.yscale("log")
    plt.xlabel("d [-]")
    plt.ylabel("Execution time [s]")

    plt.show()



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax
import jax.numpy as jnp

import piquasso as pq


def test_Beamsplitter_fock_probabilities_gradient_2_particles():
    theta = jnp.pi / 3

    simulator = pq.PureFockSimulator(
        d=2, config=pq.Config(cutoff=3), connector=pq.JaxConnector()
    )

    def func(theta):
        with pq.Program() as program:
            pq.Q(all) | pq.StateVector((2, 0))

            pq.Q(all) | pq.Beamsplitter(theta=theta)

        state = simulator.execute(program).state

        return state.fock_probabilities

    get_jacobian = jax.jacrev(func)

    fock_probabilities = func(theta)

    assert jnp.allclose(
        fock_probabilities,
        jnp.array(
            [
                0,
                0,
                0,
                jnp.cos(theta) ** 4,
                2 * (jnp.cos(theta) * jnp.sin(theta)) ** 2,
                jnp.sin(theta) ** 4,
            ]
        ),
    )

    jacobian = get_jacobian(theta)

    assert jnp.allclose(
        jacobian,
        jnp.array(
            [
                0,
                0,
                0,
                -4 * jnp.cos(theta) ** 3 * jnp.sin(theta),
                2 * (jnp.sin(2 * theta)) * jnp.cos(2 * theta),
                4 * jnp.sin(theta) ** 3 * jnp.cos(theta),
            ]
        ),
    )


def test_Beamsplitter_fock_probabilities_gradient_2_particles_JIT():
    theta = jnp.pi / 3

    simulator = pq.PureFockSimulator(
        d=2, config=pq.Config(cutoff=3), connector=pq.JaxConnector()
    )

    @jax.jit
    def func(theta):
        with pq.Program() as program:
            pq.Q(all) | pq.StateVector((2, 0))

            pq.Q(all) | pq.Beamsplitter(theta=theta)

        state = simulator.execute(program).state

        return state.fock_probabilities

    get_jacobian = jax.jacrev(func)

    fock_probabilities = func(theta)

    assert jnp.allclose(
        fock_probabilities,
        jnp.array(
            [
                0,
                0,
                0,
                jnp.cos(theta) ** 4,
                2 * (jnp.cos(theta) * jnp.sin(theta)) ** 2,
                jnp.sin(theta) ** 4,
            ]
        ),
    )

    jacobian = get_jacobian(theta)

    assert jnp.allclose(
        jacobian,
        jnp.array(
            [
                0,
                0,
                0,
                -4 * jnp.cos(theta) ** 3 * jnp.sin(theta),
                2 * (jnp.sin(2 * theta)) * jnp.cos(2 * theta),
                4 * jnp.sin(theta) ** 3 * jnp.cos(theta),
            ]
        ),
    )



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from scipy.special import factorial

from functools import partial


from piquasso._math.combinatorics import partitions

"""
This is a contribution from `theboss`, see https://github.com/Tomev-CTP/theboss.

The original code has been re-implemented and adapted to Piquasso.
"""

__author__ = "Tomasz Rybotycki"


def calculate_state_vector(interferometer, initial_state, config, connector):
    """Calculate the state vector on the particle subspace defined by ``initial_state``.

    This implementation follows Algorithm 1 (``SLOS_full``) from
    `Strong Simulation of Linear Optical Processes`.
    """

    np = connector.np
    fallback_np = connector.fallback_np

    d = len(initial_state)
    n = int(fallback_np.sum(initial_state))

    bases = [partitions(boxes=d, particles=k) for k in range(n + 1)]
    index_maps = [{tuple(basis[i]): i for i in range(len(basis))} for basis in bases]

    sigma = fallback_np.zeros(d, dtype=int)
    schedule_sigma = []
    schedule_mode = []
    for p in range(d):
        for _ in range(initial_state[p]):
            schedule_sigma.append(sigma.copy())
            schedule_mode.append(p)
            sigma[p] += 1

    UF_curr = np.zeros(len(bases[0]), dtype=config.complex_dtype)
    UF_curr = connector.assign(UF_curr, 0, 1.0)

    for k in range(n):
        sigma_k = schedule_sigma[k]
        p = schedule_mode[k]
        UF_next = np.zeros(len(bases[k + 1]), dtype=config.complex_dtype)

        for idx_t, t in enumerate(bases[k]):
            A = UF_curr[idx_t]

            for i in range(d):
                t_i = t[i]
                new_t = t.copy()
                new_t[i] += 1
                index_new = index_maps[k + 1][tuple(new_t)]

                factor = fallback_np.sqrt((t_i + 1) / (sigma_k[p] + 1))

                UF_next = connector.assign(
                    UF_next,
                    index_new,
                    UF_next[index_new] + factor * interferometer[i, p] * A,
                )

        UF_curr = UF_next

    return UF_curr


def calculate_inner_product(interferometer, input, output, connector):
    np = connector.np
    fallback_np = connector.fallback_np
    permanent = connector.permanent(interferometer, cols=input, rows=output)

    return permanent / np.sqrt(
        fallback_np.prod(factorial(output)) * fallback_np.prod(factorial(input))
    )


def generate_lossless_samples(input, shots, permanent, interferometer, rng):
    """
    Generates samples corresponding to the Clifford & Clifford algorithm B from
    https://arxiv.org/abs/1706.01260.

    Args:
        input: The input Fock basis state.
        shots: Number of samples to be generated.
        permanent: Permanent calculator function, already containing
            the unitary matrix corresponding to the interferometer.
        rng: Random number generator.

    Returns:
        The generated samples.
    """

    return _generate_samples(
        input,
        shots,
        permanent,
        sample_generator=_generate_lossless_sample,
        interferometer=interferometer,
        rng=rng,
    )


def generate_uniform_lossy_samples(
    input, shots, permanent, interferometer, transmissivity, rng
):
    """
    Basically the same algorithm as in `generate_lossless_samples`, but rejects
    particles according to the uniform `transmissivity` specified.
    """
    sample_generator = partial(
        _generate_uniform_lossy_sample, transmissivity=transmissivity
    )

    return _generate_samples(
        input,
        shots,
        permanent,
        interferometer=interferometer,
        sample_generator=sample_generator,
        rng=rng,
    )


def generate_lossy_samples(
    input_state, samples_number, permanent, interferometer_svd, rng
):
    """
    Basically the same algorithm as in `generate_lossless_samples`, but doubles the
    system size and embeds the input state and the input matrix in order to simulate
    non-uniform losses characterized by the singular values in `interferometer_svd`.
    """
    expanded_matrix = _prepare_interferometer_matrix_in_expanded_space(
        interferometer_svd
    )

    expansion_zeros = np.zeros_like(input_state, dtype=int)
    expanded_state = np.vstack([input_state, expansion_zeros])
    expanded_state = expanded_state.reshape(
        2 * len(input_state),
    )

    expanded_samples = generate_lossless_samples(
        expanded_state, samples_number, permanent, expanded_matrix, rng
    )

    # Trim output state
    samples = []

    for output_state in expanded_samples:
        while len(output_state) > len(input_state):
            output_state = np.delete(output_state, len(output_state) - 1)
        samples.append(output_state)

    return samples


def _get_first_quantized(occupation_numbers):
    first_quantized = np.empty(sum(occupation_numbers), dtype=occupation_numbers.dtype)

    idx = 0
    for mode, reps in enumerate(occupation_numbers):
        for _ in range(reps):
            first_quantized[idx] = mode
            idx += 1

    return first_quantized


def _generate_samples(input, shots, permanent, interferometer, sample_generator, rng):
    d = len(input)
    n = np.sum(input)

    samples = []

    first_quantized_input = _get_first_quantized(input)

    while len(samples) < shots:
        sample = sample_generator(
            d,
            n,
            permanent,
            interferometer,
            first_quantized_input,
            rng=rng,
        )
        samples.append(sample)

    return samples


def _grow_current_input(current_input, to_shrink, rng):
    random_index = rng.choice(len(to_shrink))
    random_mode = to_shrink[random_index]

    current_input[random_mode] += 1

    to_shrink = np.delete(to_shrink, random_index)

    return current_input, to_shrink


def _generate_lossless_sample(
    d, n, permanent, interferometer, first_quantized_input, rng
):
    sample = np.zeros(d, dtype=int)

    current_input = np.zeros(d, dtype=int)
    to_shrink = np.copy(first_quantized_input)

    for _ in range(1, n + 1):
        current_input, to_shrink = _grow_current_input(current_input, to_shrink, rng)

        pmf = _calculate_pmf(current_input, sample, permanent, interferometer)

        index = _sample_from_pmf(pmf, rng)

        sample[index] += 1

    return sample


def _generate_uniform_lossy_sample(
    d, n, permanent, interferometer, first_quantized_input, transmissivity, rng
):
    sample = np.zeros(d, dtype=int)

    number_of_particle_to_sample = 0

    current_input = np.zeros(d, dtype=int)
    to_shrink = np.copy(first_quantized_input)

    for _ in range(1, n + 1):
        if rng.uniform() > transmissivity:
            continue
        else:
            number_of_particle_to_sample += 1

        current_input, to_shrink = _grow_current_input(current_input, to_shrink, rng)

        pmf = _calculate_pmf(current_input, sample, permanent, interferometer)

        index = _sample_from_pmf(pmf, rng)

        sample[index] += 1

    return sample


def _filter_zeros(matrix, input_state, output_state):
    nonzero_input_indices = input_state > 0
    nonzero_output_indices = output_state > 0

    new_input_state = input_state[input_state > 0]
    new_output_state = output_state[output_state > 0]

    new_matrix = matrix[np.ix_(nonzero_output_indices, nonzero_input_indices)]

    return new_matrix, new_input_state, new_output_state


def _calculate_pmf(input_state, sample, calculate_permanent, interferometer):
    d = len(sample)
    pmf = np.empty(d, dtype=np.float64)

    nonzero_indices = np.arange(d)[input_state > 0]

    partial_permanents = np.zeros(len(nonzero_indices), dtype=interferometer.dtype)

    for idx, nonzero_idx in enumerate(nonzero_indices):
        input_state_subtracted = input_state.copy()
        input_state_subtracted[nonzero_idx] -= 1

        filtered_interferometer, filtered_input_state, filtered_output_state = (
            _filter_zeros(interferometer, input_state_subtracted, sample)
        )
        partial_permanents[idx] = calculate_permanent(
            filtered_interferometer, filtered_output_state, filtered_input_state
        )

    normalization = 0.0
    for idx in range(d):
        permanent = 0.0
        for j in range(len(partial_permanents)):
            permanent += (
                input_state[nonzero_indices[j]]
                * partial_permanents[j]
                * interferometer[idx, nonzero_indices[j]]
            )

        pmf[idx] = np.abs(permanent) ** 2
        normalization += pmf[idx]

    return pmf / normalization


def _sample_from_pmf(pmf, rng):
    return rng.choice(np.arange(pmf.shape[0]), p=pmf)


def _prepare_interferometer_matrix_in_expanded_space(interferometer_svd):
    v_matrix, singular_values, u_matrix = interferometer_svd

    expansions_zeros = np.zeros_like(v_matrix)
    expansions_ones = np.eye(len(v_matrix))
    expanded_v = np.block(
        [[v_matrix, expansions_zeros], [expansions_zeros, expansions_ones]]
    )
    expanded_u = np.block(
        [[u_matrix, expansions_zeros], [expansions_zeros, expansions_ones]]
    )
    singular_values_matrix_expansion = _calculate_singular_values_matrix_expansion(
        singular_values
    )
    singular_values_expanded_matrix = np.block(
        [
            [np.diag(singular_values), singular_values_matrix_expansion],
            [singular_values_matrix_expansion, np.diag(singular_values)],
        ]
    )
    return expanded_v @ singular_values_expanded_matrix @ expanded_u


def _calculate_singular_values_matrix_expansion(singular_values_vector):
    vector_of_squared_expansions = 1.0 - np.power(singular_values_vector, 2)

    expansion_values = np.sqrt(vector_of_squared_expansions)

    return np.diag(expansion_values)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from jax import jit, lax

import jax.numpy as jnp

from piquasso._math.linalg import assym_reduce


def permanent_with_reduction(matrix, rows, cols):
    """JAX implementation of the permanent, using the Glynn formula with Gray codes.

    Note:
        Unfortunately, JAX (or rather XLA) could not JIT-compile creating arrays
        whose shapes depend on the values of the input, so the reduction of the
        interferometer is needed to be made in advance.
    """

    reduced_matrix = assym_reduce(matrix, rows, cols)

    return permanent(reduced_matrix)


@jit
def permanent(M):
    n = M.shape[0]

    binary_powers = 2 ** jnp.arange(n)

    N = 2 ** (n - 1)

    def body_fun(index, val):
        sum_, sign, old_gray_code, row_sum = val

        sum_ += sign * jnp.prod(row_sum)

        new_gray_code = index ^ (index // 2)
        grey_diff = old_gray_code ^ new_gray_code
        matrix_row_index = jnp.where(binary_powers == grey_diff, size=1)[0][0]

        diff = lax.cond(
            old_gray_code > new_gray_code,
            lambda: 1,
            lambda: -1,
        )

        row_sum += M[matrix_row_index] * 2 * diff

        sign = -sign
        old_gray_code = new_gray_code

        return sum_, sign, old_gray_code, row_sum

    sum_ = 0.0
    sign = +1
    old_gray_code = 0
    row_sum = jnp.sum(M, axis=0)

    sum_ = lax.fori_loop(
        lower=1,
        upper=N + 1,
        body_fun=body_fun,
        init_val=(sum_, sign, old_gray_code, row_sum),
    )[0]

    return sum_ / N



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import piquasso as pq

from piquasso._math.combinatorics import partitions

import numpy as np


if __name__ == "__main__":
    unitary = np.array(
        [
            [
                0.90939407 + 0.26443525j,
                0.00450261 + 0.01880791j,
                0.31670376 + 0.0490014j,
            ],
            [
                0.19639913 + 0.19652978j,
                0.23847303 + 0.39887995j,
                -0.67825584 - 0.49678752j,
            ],
            [
                -0.10943493 - 0.11791454j,
                0.35197974 + 0.81225714j,
                0.3122759 + 0.30488117j,
            ],
        ]
    )

    input_state = np.array([3, 1, 1], dtype=int)

    loss_transmittance = 1.0

    n = sum(input_state)
    d = len(input_state)

    N = 10000

    program = pq.Program(
        instructions=[
            pq.StateVector(input_state),
            pq.Interferometer(unitary),
            *[pq.Loss(loss_transmittance).on_modes(i) for i in range(d)],
            pq.ParticleNumberMeasurement(),
        ]
    )

    simulator = pq.SamplingSimulator(d=len(unitary), config=pq.Config(seed_sequence=42))

    samples = [tuple(sample) for sample in simulator.execute(program, shots=N).samples]

    for partition in partitions(d, n):
        print(partition, samples.count(tuple(partition)))



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import piquasso as pq

import numpy as np
from scipy.stats import unitary_group


@pytest.fixture
def generate_unitary_matrix():
    def func(N):
        if N == 1:
            return np.array([[np.exp(np.random.rand() * 2 * np.pi * 1j)]])

        return np.array(unitary_group.rvs(N), dtype=complex)

    return func


@pytest.fixture
def d():
    return 5


@pytest.fixture
def cutoff():
    return 4


@pytest.fixture
def measurement_cutoff():
    """
    NOTE: In SF the measurement cutoff is 5, and couldn't be changed
    """
    return 5


@pytest.fixture
def example_gaussian_pq_program(d):
    with pq.Program() as program:
        for i in range(d):
            pq.Q(i) | pq.Squeezing(r=0.1) | pq.Displacement(r=1)

        pq.Q(0, 1) | pq.Beamsplitter(0.0959408065906761, 0.06786053071484363)
        pq.Q(2, 3) | pq.Beamsplitter(0.7730047654405018, 1.453770233324797)
        pq.Q(1, 2) | pq.Beamsplitter(1.0152680371119776, 1.2863559998816205)
        pq.Q(3, 4) | pq.Beamsplitter(1.3205517879465705, 0.5236836466492961)
        pq.Q(0, 1) | pq.Beamsplitter(4.394480318177715, 4.481575657714487)
        pq.Q(2, 3) | pq.Beamsplitter(2.2300919706807534, 1.5073556513699888)
        pq.Q(1, 2) | pq.Beamsplitter(2.2679037068773673, 1.9550229282085838)
        pq.Q(3, 4) | pq.Beamsplitter(3.340269832485504, 3.289367083610399)

    return program


@pytest.fixture
def example_purefock_pq_program(d):
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        for i in range(d):
            pq.Q(i) | pq.Squeezing(r=0.1) | pq.Displacement(r=1)

        pq.Q(0, 1) | pq.Beamsplitter(0.0959408065906761, 0.06786053071484363)
        pq.Q(2, 3) | pq.Beamsplitter(0.7730047654405018, 1.453770233324797)
        pq.Q(1, 2) | pq.Beamsplitter(1.0152680371119776, 1.2863559998816205)
        pq.Q(3, 4) | pq.Beamsplitter(1.3205517879465705, 0.5236836466492961)
        pq.Q(0, 1) | pq.Beamsplitter(4.394480318177715, 4.481575657714487)
        pq.Q(2, 3) | pq.Beamsplitter(2.2300919706807534, 1.5073556513699888)
        pq.Q(1, 2) | pq.Beamsplitter(2.2679037068773673, 1.9550229282085838)
        pq.Q(3, 4) | pq.Beamsplitter(3.340269832485504, 3.289367083610399)

    return program


@pytest.fixture
def pq_config(cutoff, measurement_cutoff):
    return pq.Config(cutoff=cutoff, measurement_cutoff=measurement_cutoff)


@pytest.fixture
def pq_gaussian_simulator(d, pq_config):
    return pq.GaussianSimulator(d=d, config=pq_config)


@pytest.fixture
def example_pq_gaussian_state(pq_gaussian_simulator, example_gaussian_pq_program):
    state = pq_gaussian_simulator.execute(example_gaussian_pq_program).state

    return state


@pytest.fixture
def pq_purefock_simulator(d, pq_config):
    return pq.PureFockSimulator(d=d, config=pq_config)


@pytest.fixture
def example_pq_purefock_state(pq_purefock_simulator, example_purefock_pq_program):
    state = pq_purefock_simulator.execute(example_purefock_pq_program).state

    return state


@pytest.fixture
def sf():
    try:
        import strawberryfields
    except ImportError as error:
        known_error_message = "cannot import name 'simps' from 'scipy.integrate'"

        if known_error_message in error.msg:
            raise ImportError(
                "Install a previous version of 'scipy', since 'strawberryfields' is "
                "incompatible with newer versions of scipy."
            )
        else:
            raise error

    return strawberryfields


@pytest.fixture
def example_gaussian_sf_program_and_engine(d, sf):
    """
    NOTE: the covariance matrix SF is returning is half of ours...
    It seems that our implementation is OK, however.
    """

    program = sf.Program(d)
    engine = sf.Engine(backend="gaussian")

    with program.context as q:
        sf.ops.Sgate(0.1) | q[0]
        sf.ops.Sgate(0.1) | q[1]
        sf.ops.Sgate(0.1) | q[2]
        sf.ops.Sgate(0.1) | q[3]
        sf.ops.Sgate(0.1) | q[4]

        sf.ops.Dgate(1) | q[0]
        sf.ops.Dgate(1) | q[1]
        sf.ops.Dgate(1) | q[2]
        sf.ops.Dgate(1) | q[3]
        sf.ops.Dgate(1) | q[4]

        sf.ops.BSgate(0.0959408065906761, 0.06786053071484363) | (q[0], q[1])
        sf.ops.BSgate(0.7730047654405018, 1.453770233324797) | (q[2], q[3])
        sf.ops.BSgate(1.0152680371119776, 1.2863559998816205) | (q[1], q[2])
        sf.ops.BSgate(1.3205517879465705, 0.5236836466492961) | (q[3], q[4])
        sf.ops.BSgate(4.394480318177715, 4.481575657714487) | (q[0], q[1])
        sf.ops.BSgate(2.2300919706807534, 1.5073556513699888) | (q[2], q[3])
        sf.ops.BSgate(2.2679037068773673, 1.9550229282085838) | (q[1], q[2])
        sf.ops.BSgate(3.340269832485504, 3.289367083610399) | (q[3], q[4])

    return program, engine


@pytest.fixture
def example_fock_sf_program_and_engine(d, cutoff, sf):
    """
    NOTE: the covariance matrix SF is returning is half of ours...
    It seems that our implementation is OK, however.
    """

    program = sf.Program(d)
    engine = sf.Engine(backend="fock", backend_options={"cutoff_dim": cutoff})

    with program.context as q:
        sf.ops.Sgate(0.1) | q[0]
        sf.ops.Sgate(0.1) | q[1]
        sf.ops.Sgate(0.1) | q[2]
        sf.ops.Sgate(0.1) | q[3]
        sf.ops.Sgate(0.1) | q[4]

        sf.ops.Dgate(1) | q[0]
        sf.ops.Dgate(1) | q[1]
        sf.ops.Dgate(1) | q[2]
        sf.ops.Dgate(1) | q[3]
        sf.ops.Dgate(1) | q[4]

        sf.ops.BSgate(0.0959408065906761, 0.06786053071484363) | (q[0], q[1])
        sf.ops.BSgate(0.7730047654405018, 1.453770233324797) | (q[2], q[3])
        sf.ops.BSgate(1.0152680371119776, 1.2863559998816205) | (q[1], q[2])
        sf.ops.BSgate(1.3205517879465705, 0.5236836466492961) | (q[3], q[4])
        sf.ops.BSgate(4.394480318177715, 4.481575657714487) | (q[0], q[1])
        sf.ops.BSgate(2.2300919706807534, 1.5073556513699888) | (q[2], q[3])
        sf.ops.BSgate(2.2679037068773673, 1.9550229282085838) | (q[1], q[2])
        sf.ops.BSgate(3.340269832485504, 3.289367083610399) | (q[3], q[4])

    return program, engine


@pytest.fixture
def example_sf_gaussian_state(example_gaussian_sf_program_and_engine):
    program, engine = example_gaussian_sf_program_and_engine
    results = engine.run(program)

    return results.state


@pytest.fixture
def example_sf_fock_state(example_fock_sf_program_and_engine):
    program, engine = example_fock_sf_program_and_engine
    results = engine.run(program)

    return results.state



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import piquasso as pq

from piquasso.decompositions.clements import clements, instructions_from_decomposition

import numpy as np

import pytest

for_all_connectors = pytest.mark.parametrize(
    "connector", [pq.NumpyConnector(), pq.JaxConnector()]
)


@for_all_connectors
def test_Interferometer_2_by_2(connector):
    theta = np.pi / 5

    U = np.array([[np.cos(theta), -np.sin(theta)], [np.sin(theta), np.cos(theta)]])

    with pq.Program() as program:
        pq.Q(0, 1) | pq.StateVector([0, 1])
        pq.Q(0, 1) | pq.Interferometer(U)

    fock_simulator = pq.fermionic.PureFockSimulator(d=2, connector=connector)

    fock_state = fock_simulator.execute(program).state

    assert np.allclose(fock_state.state_vector, [0, -np.sin(theta), np.cos(theta), 0])


@pytest.mark.monkey
@for_all_connectors
def test_Interferometer_3_by_3_random(connector, generate_unitary_matrix):
    d = 3
    U = generate_unitary_matrix(d)

    with pq.Program() as program:
        # NOTE: This violates the parity superselection rule, but the simulator should
        # permit it.
        pq.Q(0, 1, 2) | pq.StateVector([0, 0, 1]) / np.sqrt(2)
        pq.Q(0, 1, 2) | pq.StateVector([0, 1, 1]) / np.sqrt(2)
        pq.Q(0, 1, 2) | pq.Interferometer(U)

    fock_simulator = pq.fermionic.PureFockSimulator(d=d, connector=connector)

    fock_state = fock_simulator.execute(program).state

    assert np.allclose(
        fock_state.state_vector,
        1
        / np.sqrt(2)
        * np.array(
            [
                0.0,
                U[0, 2],
                U[1, 2],
                U[2, 2],
                U[0, 1] * U[1, 2] - U[0, 2] * U[1, 1],
                U[0, 1] * U[2, 2] - U[0, 2] * U[2, 1],
                U[1, 1] * U[2, 2] - U[1, 2] * U[2, 1],
                0.0,
            ]
        ),
    )


@pytest.mark.monkey
@for_all_connectors
def test_Interferometer_clements_equivalence(connector, generate_unitary_matrix):
    d = 3

    U = generate_unitary_matrix(d)

    with pq.Program() as preparation:
        pq.Q() | pq.StateVector(occupation_numbers=[1, 0, 1]) / np.sqrt(2)
        pq.Q() | pq.StateVector(occupation_numbers=[0, 1, 1]) / np.sqrt(2)

    with pq.Program() as program:
        pq.Q() | preparation
        pq.Q() | pq.Interferometer(matrix=U)

    decomposition = clements(U, connector=connector)

    with pq.Program() as decomposed_program:
        pq.Q() | preparation
        decomposed_program.instructions.extend(
            instructions_from_decomposition(decomposition)
        )

    simulator = pq.fermionic.PureFockSimulator(d=d, connector=connector)

    state = simulator.execute(program).state
    decomposed_state = simulator.execute(decomposed_program).state

    assert state == decomposed_state


@pytest.mark.monkey
@for_all_connectors
def test_Interferometer_subsystem_equivalence(connector, generate_unitary_matrix):
    d = 3

    U = generate_unitary_matrix(2)

    bigU = np.identity(d, dtype=complex)

    bigU[0, 0] = U[0, 0]
    bigU[0, 1] = U[0, 1]
    bigU[1, 0] = U[1, 0]
    bigU[1, 1] = U[1, 1]

    with pq.Program() as preparation:
        pq.Q() | pq.StateVector(occupation_numbers=[0, 0, 1]) / np.sqrt(3)
        pq.Q() | pq.StateVector(occupation_numbers=[1, 0, 1]) / np.sqrt(3)
        pq.Q() | pq.StateVector(occupation_numbers=[0, 1, 1]) / np.sqrt(3)

    with pq.Program() as program_subsystem:
        pq.Q() | preparation
        pq.Q(0, 1) | pq.Interferometer(matrix=U)

    with pq.Program() as program_full:
        pq.Q() | preparation
        pq.Q(0, 1, 2) | pq.Interferometer(matrix=bigU)

    simulator = pq.fermionic.PureFockSimulator(d=d, connector=connector)

    state_subsystem = simulator.execute(program_subsystem).state
    state_full = simulator.execute(program_full).state

    assert state_subsystem == state_full


@for_all_connectors
def test_Interferometer_nonconsecutive_ordering_raises_InvalidParameter(connector):
    d = 3

    self_adjoint = np.array(
        [
            [1, 2j, 3 + 4j],
            [-2j, 2, 5],
            [3 - 4j, 5, 6],
        ],
        dtype=complex,
    )
    unitary = connector.expm(1j * self_adjoint)

    simulator = pq.fermionic.PureFockSimulator(d=d, connector=connector)

    with pq.Program() as preparation:
        pq.Q(0, 1, 2) | pq.StateVector([1, 1, 0])
        pq.Q(0, 2, 1) | pq.Interferometer(unitary)

    with pytest.raises(pq.api.exceptions.InvalidParameter) as error:
        simulator.execute(preparation)

    assert error.value.args[0] == "Specified modes must be consecutive: modes=(0, 2, 1)"


@for_all_connectors
def test_Squeezing2_on_two_modes(connector):
    d = 2

    r = 0.1
    phi = np.pi / 7

    with pq.Program() as program:
        pq.Q() | pq.StateVector([1, 1])

        pq.Q(0, 1) | pq.Squeezing2(r=r, phi=phi)

    simulator = pq.fermionic.PureFockSimulator(
        d=d, config=pq.Config(cutoff=d + 1), connector=connector
    )

    state = simulator.execute(program).state

    term_00 = np.sin(r / 2) * np.exp(-1j * phi)
    term_11 = np.cos(r / 2)

    expected_state_vector = np.array([term_00, 0.0, 0.0, term_11])

    assert np.allclose(state.state_vector, expected_state_vector)


@for_all_connectors
def test_Squeezing2_nonconsecutive_ordering_raises_InvalidParameter(connector):
    d = 3

    r = 0.1
    phi = np.pi / 7

    with pq.Program() as program:
        pq.Q() | pq.StateVector([1, 0, 1])

        pq.Q(0, 2) | pq.Squeezing2(r=r, phi=phi)

    simulator = pq.fermionic.PureFockSimulator(
        d=d, config=pq.Config(cutoff=d + 1), connector=connector
    )

    with pytest.raises(pq.api.exceptions.InvalidParameter) as error:
        simulator.execute(program)

    assert error.value.args[0] == "Specified modes must be consecutive: modes=(0, 2)"


@for_all_connectors
def test_Squeezing2_cutoff(connector):
    d = 3

    r = 0.1
    phi = np.pi / 7

    cutoff = 2

    with pq.Program() as program:
        pq.Q() | pq.StateVector([0, 0, 0])

        pq.Q(0, 1) | pq.Squeezing2(r=r, phi=phi)

    simulator = pq.fermionic.PureFockSimulator(
        d=d, config=pq.Config(cutoff=cutoff), connector=connector
    )

    state = simulator.execute(program).state

    expected_state_vector = np.array([np.cos(r / 2), 0.0, 0.0, 0.0])

    assert np.allclose(state.state_vector, expected_state_vector)


@for_all_connectors
def test_ControlledPhase(connector):
    d = 2
    r = 0.1

    phi = np.pi / 7
    cp_phi = np.pi / 5

    with pq.Program() as program:
        pq.Q() | pq.StateVector([1, 1])

        pq.Q(0, 1) | pq.Squeezing2(r=r, phi=phi)

        pq.Q(0, 1) | pq.fermionic.ControlledPhase(phi=cp_phi)

    simulator = pq.fermionic.PureFockSimulator(
        d=d, config=pq.Config(cutoff=d + 1), connector=connector
    )

    state = simulator.execute(program).state

    term_00 = np.sin(r / 2) * np.exp(-1j * phi)
    term_11 = np.cos(r / 2) * np.exp(1j * cp_phi)

    expected_state_vector = np.array([term_00, 0.0, 0.0, term_11])

    assert np.allclose(state.state_vector, expected_state_vector)


@for_all_connectors
def test_IsingXX(connector):
    d = 2

    phi = np.pi / 7
    a = np.sqrt(3 / 4)
    b = np.sqrt(1 / 4)

    with pq.Program() as program:
        pq.Q() | pq.StateVector([0, 0]) * a
        pq.Q() | pq.StateVector([0, 1]) * b

        pq.Q(0, 1) | pq.fermionic.IsingXX(phi=phi)

    simulator = pq.fermionic.PureFockSimulator(
        d=d, config=pq.Config(cutoff=d + 1), connector=connector
    )

    state = simulator.execute(program).state

    expected_state_vector = np.array(
        [
            a * np.cos(phi),
            b * 1j * np.sin(phi),
            b * np.cos(phi),
            a * 1j * np.sin(phi),
        ]
    )

    assert np.allclose(state.state_vector, expected_state_vector)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

import numpy as np

import piquasso as pq


def test_BatchPureFockState_equals_to_itself():
    with pq.Program() as first_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(1) | pq.Squeezing(r=0.1, phi=np.pi / 4)

    with pq.Program() as second_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(0) | pq.Displacement(r=1.0, phi=np.pi / 10)

    with pq.Program() as batch_program:
        pq.Q() | pq.BatchPrepare([first_preparation, second_preparation])

        pq.Q() | pq.Beamsplitter(theta=np.pi / 6, phi=np.pi / 3)

    simulator = pq.PureFockSimulator(d=2, config=pq.Config(cutoff=5))

    batch_state = simulator.execute(batch_program).state

    assert batch_state == batch_state


def test_BatchPureFockState_without_normalization():
    with pq.Program() as first_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(1) | pq.Squeezing(r=0.1, phi=np.pi / 4)

    with pq.Program() as second_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(0) | pq.Displacement(r=1.0, phi=np.pi / 10)

    with pq.Program() as batch_program:
        pq.Q() | pq.BatchPrepare([first_preparation, second_preparation])

        pq.Q() | pq.Beamsplitter(theta=np.pi / 6, phi=np.pi / 3)

    simulator = pq.PureFockSimulator(d=2, config=pq.Config(cutoff=5))

    batch_state = simulator.execute(batch_program).state

    with pytest.raises(pq.api.exceptions.InvalidState) as error:
        batch_state.validate()

    assert error.value.args[0] == (
        "The sum of probabilities is not close to 1.0 for at least one state in the "
        "batch."
    )


def test_BatchPureFockState_without_normalization_but_validate_False():
    with pq.Program() as first_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(1) | pq.Squeezing(r=0.1, phi=np.pi / 4)

    with pq.Program() as second_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(0) | pq.Displacement(r=1.0, phi=np.pi / 10)

    with pq.Program() as batch_program:
        pq.Q() | pq.BatchPrepare([first_preparation, second_preparation])

        pq.Q() | pq.Beamsplitter(theta=np.pi / 6, phi=np.pi / 3)

    simulator = pq.PureFockSimulator(d=2, config=pq.Config(cutoff=5, validate=False))

    batch_state = simulator.execute(batch_program).state

    batch_state.validate()


def test_invalid_BatchPureFockState_normalize():
    with pq.Program() as first_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(1) | pq.Squeezing(r=0.1, phi=np.pi / 4)

    with pq.Program() as invalid_second_preparation:
        pass

    with pq.Program() as batch_program:
        pq.Q() | pq.BatchPrepare([first_preparation, invalid_second_preparation])

        pq.Q() | pq.Beamsplitter(theta=np.pi / 6, phi=np.pi / 3)

    simulator = pq.PureFockSimulator(d=2, config=pq.Config(cutoff=5))

    batch_state = simulator.execute(batch_program).state

    with pytest.raises(pq.api.exceptions.InvalidState) as error:
        batch_state.normalize()

    assert "The norm of a state in the batch is 0." == error.value.args[0]


def test_batch_Beamsplitter_state_vector():
    with pq.Program() as first_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(1) | pq.Squeezing(r=0.1, phi=np.pi / 4)

    with pq.Program() as second_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(0) | pq.Displacement(r=1.0, phi=np.pi / 10)

    with pq.Program() as common:
        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 6, phi=np.pi / 3)

    with pq.Program() as batch_program:
        pq.Q() | pq.BatchPrepare([first_preparation, second_preparation])

        pq.Q() | common

    with pq.Program() as first_program:
        pq.Q() | first_preparation

        pq.Q() | common

    with pq.Program() as second_program:
        pq.Q() | second_preparation

        pq.Q() | common

    simulator = pq.PureFockSimulator(d=2, config=pq.Config(cutoff=5))

    batch_state_vector = simulator.execute(batch_program).state.state_vector

    first_state_vector = simulator.execute(first_program).state.state_vector
    second_state_vector = simulator.execute(second_program).state.state_vector

    assert np.allclose(batch_state_vector[:, 0], first_state_vector)
    assert np.allclose(batch_state_vector[:, 1], second_state_vector)


def test_batch_Squeezing_and_Displacement_state_vector():
    with pq.Program() as first_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(1) | pq.Squeezing(r=0.1, phi=np.pi / 4)

    with pq.Program() as second_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(0) | pq.Displacement(r=1.0, phi=np.pi / 10)

    with pq.Program() as common:
        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 6, phi=np.pi / 3)

        pq.Q(0) | pq.Displacement(1.0)
        pq.Q(0) | pq.Squeezing(0.1)

    with pq.Program() as batch_program:
        pq.Q() | pq.BatchPrepare([first_preparation, second_preparation])

        pq.Q() | common

    with pq.Program() as first_program:
        pq.Q() | first_preparation

        pq.Q() | common

    with pq.Program() as second_program:
        pq.Q() | second_preparation

        pq.Q() | common

    simulator = pq.PureFockSimulator(d=2, config=pq.Config(cutoff=5))

    batch_state_vector = simulator.execute(batch_program).state.state_vector

    first_state_vector = simulator.execute(first_program).state.state_vector
    second_state_vector = simulator.execute(second_program).state.state_vector

    assert np.allclose(batch_state_vector[:, 0], first_state_vector)
    assert np.allclose(batch_state_vector[:, 1], second_state_vector)


def test_batch_Kerr_state_vector():
    with pq.Program() as first_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(1) | pq.Squeezing(r=0.1, phi=np.pi / 4)

    with pq.Program() as second_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(0) | pq.Displacement(r=1.0, phi=np.pi / 10)

    with pq.Program() as common:
        pq.Q(0) | pq.Kerr(0.1)
        pq.Q(1) | pq.Kerr(0.2)

    with pq.Program() as batch_program:
        pq.Q() | pq.BatchPrepare([first_preparation, second_preparation])

        pq.Q() | common

    with pq.Program() as first_program:
        pq.Q() | first_preparation

        pq.Q() | common

    with pq.Program() as second_program:
        pq.Q() | second_preparation

        pq.Q() | common

    simulator = pq.PureFockSimulator(d=2, config=pq.Config(cutoff=5))

    batch_state_vector = simulator.execute(batch_program).state.state_vector

    first_state_vector = simulator.execute(first_program).state.state_vector
    second_state_vector = simulator.execute(second_program).state.state_vector

    assert np.allclose(batch_state_vector[:, 0], first_state_vector)
    assert np.allclose(batch_state_vector[:, 1], second_state_vector)


def test_batch_Phaseshifter_state_vector():
    with pq.Program() as first_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(1) | pq.Squeezing(r=0.1, phi=np.pi / 4)

    with pq.Program() as second_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(0) | pq.Displacement(r=1.0, phi=np.pi / 10)

    with pq.Program() as common:
        pq.Q(0) | pq.Phaseshifter(0.1)
        pq.Q(1) | pq.Phaseshifter(0.2)

    with pq.Program() as batch_program:
        pq.Q() | pq.BatchPrepare([first_preparation, second_preparation])

        pq.Q() | common

    with pq.Program() as first_program:
        pq.Q() | first_preparation

        pq.Q() | common

    with pq.Program() as second_program:
        pq.Q() | second_preparation

        pq.Q() | common

    simulator = pq.PureFockSimulator(d=2, config=pq.Config(cutoff=5))

    batch_state_vector = simulator.execute(batch_program).state.state_vector

    first_state_vector = simulator.execute(first_program).state.state_vector
    second_state_vector = simulator.execute(second_program).state.state_vector

    assert np.allclose(batch_state_vector[:, 0], first_state_vector)
    assert np.allclose(batch_state_vector[:, 1], second_state_vector)


def test_batch_multiple_gates_state_vector():
    with pq.Program() as first_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(1) | pq.Squeezing(r=0.1, phi=np.pi / 4)

    with pq.Program() as second_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(0) | pq.Displacement(r=1.0, phi=np.pi / 10)

    with pq.Program() as common:
        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 6, phi=np.pi / 3)

        pq.Q(0) | pq.Phaseshifter(0.1)
        pq.Q(1) | pq.Phaseshifter(0.2)

        pq.Q(0) | pq.Squeezing(0.1)
        pq.Q(0) | pq.Squeezing(0.2)

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 5, phi=np.pi / 7)

        pq.Q(0) | pq.Phaseshifter(0.5)
        pq.Q(1) | pq.Phaseshifter(0.2)

        pq.Q(0) | pq.Displacement(1.0)
        pq.Q(0) | pq.Displacement(1.0, phi=np.pi / 2)

        pq.Q(0) | pq.Kerr(0.1)
        pq.Q(1) | pq.Kerr(0.2)

    with pq.Program() as batch_program:
        pq.Q() | pq.BatchPrepare([first_preparation, second_preparation])

        pq.Q() | common

    with pq.Program() as first_program:
        pq.Q() | first_preparation

        pq.Q() | common

    with pq.Program() as second_program:
        pq.Q() | second_preparation

        pq.Q() | common

    simulator = pq.PureFockSimulator(d=2, config=pq.Config(cutoff=5))

    batch_state_vector = simulator.execute(batch_program).state.state_vector

    first_state_vector = simulator.execute(first_program).state.state_vector
    second_state_vector = simulator.execute(second_program).state.state_vector

    assert np.allclose(batch_state_vector[:, 0], first_state_vector)
    assert np.allclose(batch_state_vector[:, 1], second_state_vector)


def test_batch_multiple_gates_nonzero_elements():
    with pq.Program() as first_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(1) | pq.Squeezing(r=0.1, phi=np.pi / 4)

    with pq.Program() as second_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(0) | pq.Displacement(r=1.0, phi=np.pi / 10)

    with pq.Program() as common:
        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 6, phi=np.pi / 3)

        pq.Q(0) | pq.Phaseshifter(0.1)
        pq.Q(1) | pq.Phaseshifter(0.2)

        pq.Q(0) | pq.Squeezing(0.1)
        pq.Q(0) | pq.Squeezing(0.2)

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 5, phi=np.pi / 7)

        pq.Q(0) | pq.Phaseshifter(0.5)
        pq.Q(1) | pq.Phaseshifter(0.2)

        pq.Q(0) | pq.Displacement(1.0)
        pq.Q(0) | pq.Displacement(1.0, phi=np.pi / 2)

        pq.Q(0) | pq.Kerr(0.1)
        pq.Q(1) | pq.Kerr(0.2)

    with pq.Program() as batch_program:
        pq.Q() | pq.BatchPrepare([first_preparation, second_preparation])

        pq.Q() | common

    with pq.Program() as first_program:
        pq.Q() | first_preparation

        pq.Q() | common

    with pq.Program() as second_program:
        pq.Q() | second_preparation

        pq.Q() | common

    simulator = pq.PureFockSimulator(d=2, config=pq.Config(cutoff=5))

    batch_state_as_string = str(simulator.execute(batch_program).state)

    first_state_as_string = str(simulator.execute(first_program).state)
    second_state_as_string = str(simulator.execute(second_program).state)

    assert batch_state_as_string == f"{first_state_as_string}\n{second_state_as_string}"


def test_batch_mean_position():
    with pq.Program() as first_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(1) | pq.Squeezing(r=0.1, phi=np.pi / 4)

    with pq.Program() as second_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(0) | pq.Displacement(r=1.0, phi=np.pi / 10)

    with pq.Program() as common:
        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 6, phi=np.pi / 3)

        pq.Q(0) | pq.Displacement(1.0)
        pq.Q(0) | pq.Squeezing(0.1)

        pq.Q(0) | pq.Kerr(0.1)
        pq.Q(1) | pq.Kerr(0.2)

    with pq.Program() as batch_program:
        pq.Q() | pq.BatchPrepare([first_preparation, second_preparation])

        pq.Q() | common

    with pq.Program() as first_program:
        pq.Q() | first_preparation

        pq.Q() | common

    with pq.Program() as second_program:
        pq.Q() | second_preparation

        pq.Q() | common

    simulator = pq.PureFockSimulator(d=2, config=pq.Config(cutoff=5))

    batch_mean_positions = simulator.execute(batch_program).state.mean_position(0)

    first_mean_position = simulator.execute(first_program).state.mean_position(0)
    second_mean_position = simulator.execute(second_program).state.mean_position(0)

    assert np.allclose(batch_mean_positions[0], first_mean_position)
    assert np.allclose(batch_mean_positions[1], second_mean_position)


def test_Batch_with_OneByOne():
    with pq.Program() as first_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(1) | pq.Squeezing(r=0.1, phi=np.pi / 4)

    with pq.Program() as second_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(0) | pq.Displacement(r=1.0, phi=np.pi / 10)

    with pq.Program() as common1:
        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 6, phi=np.pi / 3)

        pq.Q(0) | pq.Displacement(1.0)
        pq.Q(0) | pq.Squeezing(0.1)

    with pq.Program() as common2:
        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 5, phi=np.pi / 6)

        pq.Q(0) | pq.Displacement(-1.0, phi=np.pi / 3)
        pq.Q(0) | pq.Squeezing(0.2, phi=np.pi / 5)

    with pq.Program() as first_intermediate:
        pq.Q(0) | pq.Kerr(0.1)

    with pq.Program() as second_intermediate:
        pq.Q(1) | pq.Kerr(0.2)

    with pq.Program() as batch_program:
        pq.Q() | pq.BatchPrepare([first_preparation, second_preparation])

        pq.Q() | common1

        pq.Q() | pq.BatchApply([first_intermediate, second_intermediate])

        pq.Q() | common2

    with pq.Program() as first_program:
        pq.Q() | first_preparation

        pq.Q() | common1

        pq.Q() | first_intermediate

        pq.Q() | common2

    with pq.Program() as second_program:
        pq.Q() | second_preparation

        pq.Q() | common1

        pq.Q() | second_intermediate

        pq.Q() | common2

    simulator = pq.PureFockSimulator(d=2, config=pq.Config(cutoff=5))

    batch_state_vector = simulator.execute(batch_program).state.state_vector

    first_state_vector = simulator.execute(first_program).state.state_vector
    second_state_vector = simulator.execute(second_program).state.state_vector

    assert np.allclose(batch_state_vector[:, 0], first_state_vector)
    assert np.allclose(batch_state_vector[:, 1], second_state_vector)


def test_batch_state_fock_amplitudes_map_output_type():
    with pq.Program() as first_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(1) | pq.Squeezing(r=0.1, phi=np.pi / 4)

    with pq.Program() as second_preparation:
        pq.Q() | pq.Vacuum()
        pq.Q(0) | pq.Displacement(r=1.0, phi=np.pi / 10)

    with pq.Program() as common:
        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 6, phi=np.pi / 3)

    with pq.Program() as batch_program:
        pq.Q() | pq.BatchPrepare([first_preparation, second_preparation])

        pq.Q() | common

    simulator = pq.PureFockSimulator(d=2, config=pq.Config(cutoff=5))

    batch_state = simulator.execute(batch_program).state
    assert len(batch_state.fock_amplitudes_map) == 2
    assert isinstance(batch_state.fock_amplitudes_map[0], dict)
    assert isinstance(batch_state.fock_amplitudes_map[1], dict)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This file contains some calculations for `JaxConnector`.
"""

from functools import partial

import jax
import jax.numpy as jnp

from ..connections import _calculate_index_list_for_appling_interferometer
from .._utils import precalculate_fermionic_passive_linear_indices


@partial(jax.jit, static_argnames="cutoff")
def calculate_interferometer_on_fermionic_fock_space(matrix, cutoff):
    """Calculates the representation of the interferometer matrix on the Fock space.

    This algorithm calculates the determinants of the submatrices recursively using
    Laplace's expansion. For this, it is easiest to work in the first quantized picture.

    Moreover, it is implemented so that the JAX compilation time is minimal, keeping in
    mind that the cutoff is constant during a simulation.
    """
    d = len(matrix)

    subspace_representations = []

    subspace_representations.append(jnp.array([[1.0]], dtype=matrix.dtype))

    if cutoff == 1:
        return subspace_representations

    subspace_representations.append(matrix)

    if cutoff == 2:
        return subspace_representations

    for n in range(2, cutoff):
        laplace_indices, deleted_indices = (
            precalculate_fermionic_passive_linear_indices(n, d)
        )

        matrix_row_indices = laplace_indices[:, 0]
        deleted_row_indices = deleted_indices[:, 0]

        previous_representation = subspace_representations[n - 1]

        signs = jnp.where(jnp.arange(n) % 2 == 0, 1, -1)
        matrix_term = matrix[matrix_row_indices][:, laplace_indices]
        previous_representation_term = previous_representation[deleted_row_indices][
            :, deleted_indices
        ]

        product = signs * matrix_term * previous_representation_term

        representation = jnp.sum(product, axis=2)

        subspace_representations.append(representation)

    return subspace_representations


@partial(jax.jit, static_argnames=["modes", "d", "cutoff"])
def apply_fermionic_passive_linear_to_state_vector(
    representations, state_vector, modes, d, cutoff
):
    """Applies a passive linear gate to a state vector expressed in the Fock basis.

    This function assumes that the n-particle representations of the passive linear gate
    has already been calculated.
    """
    index_list = _calculate_index_list_for_appling_interferometer(modes, d, cutoff)

    new_state_vector = jnp.zeros_like(state_vector)

    for n, indices in enumerate(index_list):
        new_state_vector = new_state_vector.at[indices].set(
            representations[n] @ state_vector[indices]
        )

    return new_state_vector



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from piquasso._math.linalg import block_reduce


def test_block_reduce_on_2_by_2_matrix():
    matrix = np.array(
        [
            [11, 12],
            [21, 22],
        ],
        dtype=float,
    )

    result = block_reduce(matrix, reduce_on=(2,))

    assert np.allclose(
        result,
        np.array(
            [
                [11, 11, 12, 12],
                [11, 11, 12, 12],
                [21, 21, 22, 22],
                [21, 21, 22, 22],
            ]
        ),
    )


def test_block_reduce_on_4_by_4_matrix():
    matrix = np.array(
        [
            [11, 12, 13, 14],
            [21, 22, 23, 24],
            [31, 32, 33, 34],
            [41, 42, 43, 44],
        ],
        dtype=float,
    )

    result = block_reduce(matrix, reduce_on=(1, 2))

    assert np.allclose(
        result,
        np.array(
            [
                [11, 12, 12, 13, 14, 14],
                [21, 22, 22, 23, 24, 24],
                [21, 22, 22, 23, 24, 24],
                [31, 32, 32, 33, 34, 34],
                [41, 42, 42, 43, 44, 44],
                [41, 42, 42, 43, 44, 44],
            ],
        ),
    )


def test_hafnian_on_6_by_6_matrix():
    matrix = np.array(
        [
            [11, 12, 13, 14, 15, 16],
            [21, 22, 23, 24, 25, 26],
            [31, 32, 33, 34, 35, 36],
            [41, 42, 43, 44, 45, 46],
            [51, 52, 53, 54, 55, 56],
            [61, 62, 63, 64, 65, 66],
        ],
        dtype=float,
    )

    result = block_reduce(matrix, reduce_on=(1, 0, 2))

    assert np.allclose(
        result,
        np.array(
            [
                [11, 13, 13, 14, 16, 16],
                [31, 33, 33, 34, 36, 36],
                [31, 33, 33, 34, 36, 36],
                [41, 43, 43, 44, 46, 46],
                [61, 63, 63, 64, 66, 66],
                [61, 63, 63, 64, 66, 66],
            ],
        ),
    )



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc

from typing import Any, Tuple

import numpy


class BaseConnector(abc.ABC):
    """The base class for encapsulating a framework.

    This class makes dependency injection into Piquasso possible. This way, Piquasso
    calculations can be made independent of the underlying framework (NumPy, TensorFlow,
    JAX) using the framework's implementation of the NumPy API.

    :ivar np: The implemented NumPy API.
    :ivar fallback_np:
        Usually the original NumPy API for calculations that are not required to be
        present on the computation graph.
    :ivar forward_pass_np: NumPy API used in the forward pass.

    See:
        - `Interoperability with NumPy <https://numpy.org/devdocs/user/basics.interoperability.html>`_

    Note:
        Every attribute of this class should be stateless!
    """  # noqa: E501

    allow_conditionals = True

    np: Any
    fallback_np: Any
    forward_pass_np: Any
    range: Any
    sqrtm: Any
    schur: Any

    def __deepcopy__(self, memo: Any) -> "BaseConnector":
        """
        This method exists, because `copy.deepcopy` could not copy the entire modules
        and functions, and we don't need to, since every attribute of this class should
        be stateless.
        """

        return self

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}()"

    @abc.abstractmethod
    def preprocess_input_for_custom_gradient(self, value):
        """
        Applies modifications to inputs in custom gradients.
        """

    @abc.abstractmethod
    def permanent(
        self, matrix: numpy.ndarray, rows: Tuple[int, ...], columns: Tuple[int, ...]
    ) -> float:
        """Calculates the permanent of a matrix with row and column repetitions."""

    @abc.abstractmethod
    def hafnian(self, matrix: numpy.ndarray, reduce_on: numpy.ndarray) -> float:
        r"""Calculates the hafnian of a matrix with prescribed reduction array.

        This function first performs a reduction by a reduction array :math:`S`, and
        then calculates the hafnian. Succintly, this function should implement
        :math:`A \mapsto \operatorname{haf}(A_{(S)})`.

        Args:
            matrix (numpy.ndarray): The input matrix.
            reduce_on (numpy.ndarray): The reduction array.

        Returns:
            float: The hafnian of the matrix.
        """

    @abc.abstractmethod
    def loop_hafnian(
        self, matrix: numpy.ndarray, diagonal: numpy.ndarray, reduce_on: numpy.ndarray
    ) -> float:
        r"""Calculates the hafnian of a matrix with prescribed reduction array.

        This function first fills the diagonals with :math:`D`, then performs a
        reduction by a reduction array :math:`S` and then calculates the hafnian.
        Succintly, this function should implement
        :math:`A \mapsto \operatorname{lhaf}(\operatorname{filldiag}(A, D)_{(S)})`.

        Args:
            matrix (numpy.ndarray): The input matrix.
            diagonal (numpy.ndarray): The vector which will fill the diagonal.
            reduce_on (numpy.ndarray): The reduction array.

        Returns:
            float: The hafnian of the matrix.
        """

    @abc.abstractmethod
    def loop_hafnian_batch(
        self,
        matrix: numpy.ndarray,
        diagonal: numpy.ndarray,
        reduce_on: numpy.ndarray,
        cutoff: int,
    ) -> float:
        r"""Batch loop hafnian calculation.

        Same as :meth:`loop_hafnian`, but with batching, according to
        https://arxiv.org/abs/2108.01622.
        """

    @abc.abstractmethod
    def assign(self, array, index, value):
        """Item assignment."""

    @abc.abstractmethod
    def scatter(self, indices, updates, shape):
        """Filling an array of a given shape with the given indices and update values.

        Equivalent to :func:`tf.scatter_nd`.
        """

    @abc.abstractmethod
    def embed_in_identity(self, matrix, indices, dim):
        """Embeds a matrix in identity."""

    @abc.abstractmethod
    def block(self, arrays):
        """Assembling submatrices into a single matrix.

        Equivalent to :func:`numpy.block`.
        """

    @abc.abstractmethod
    def block_diag(self, *arrs):
        """Putting together matrices as a block diagonal matrix.

        Equivalent to :func:`scipy.linalg.block_diag`.
        """

    @abc.abstractmethod
    def polar(self, matrix, side="right"):
        """Polar decomposition.

        Equivalent to :func:`scipy.linalg.polar`.

        Args:
            matrix (numpy.ndarray): The input matrix
            side (str, optional): The order of the decomposition. Defaults to "right".
        """

    @abc.abstractmethod
    def logm(self, matrix):
        """Matrix logarithm.

        Equivalent to :func:`scipy.linalg.logm`.

        Args:
            matrix (numpy.ndarray): The input matrix.
        """

    @abc.abstractmethod
    def expm(self, matrix):
        """Matrix exponential.

        Equivalent to :func:`scipy.linalg.expm`.

        Args:
            matrix (numpy.ndarray): The input matrix.
        """

    @abc.abstractmethod
    def powm(self, matrix, power):
        """Matrix power.

        Equivalent to :func:`numpy.linalg.matrix_power`.

        Args:
            matrix (numpy.ndarray): The input matrix.
        """

    @abc.abstractmethod
    def custom_gradient(self, func):
        """Custom gradient wrapper.

        Args:
            func: The function for which custom gradient is defined.
        """

    @abc.abstractmethod
    def accumulator(self, dtype, size, **kwargs):
        """Datatype to collect NumPy arrays.

        Common generalization of a Python list and
        `tf.TensorArray <https://www.tensorflow.org/api_docs/python/tf/TensorArray>`_.
        """

    @abc.abstractmethod
    def write_to_accumulator(self, accumulator, index, value):
        """Append an element to the accumulator.

        Common generalization of a Python list appending and
        `tf.TensorArray.write <https://www.tensorflow.org/api_docs/python/tf/TensorArray#write>`_.
        """  # noqa: E501

    @abc.abstractmethod
    def stack_accumulator(self, accumulator):
        """Stack elements in the accumulator.

        Common generalization of :func:`numpy.stack` and
        `tf.TensorArray.stack <https://www.tensorflow.org/api_docs/python/tf/TensorArray#stack>`_.
        """  # noqa: E501

    @abc.abstractmethod
    def decorator(self, func):
        """Decorates heavy computations in Piquasso.

        Args:
            func: Function to decorate.
        """

    @abc.abstractmethod
    def gather_along_axis_1(self, array, indices):
        """Gathering values along axis 1 of a matrix.

        Note:
            Gather along axis 1 was terribly slow in Tensorflow, see
            https://github.com/tensorflow/ranking/issues/160.
        """

    @abc.abstractmethod
    def transpose(self, matrix):
        """Matrix transposition.

        Args:
            matrix (numpy.ndarray): The input matrix.
        """

    @abc.abstractmethod
    def calculate_interferometer_on_fock_space(self, interferometer, helper_indices):
        """Calculates the interferometer unitary matrix on the Fock space.

        Args:
            interferometer: One-particle unitary corresponding to the interferometer.
            helper_indices: Separately calculated helper indices.

        Returns:
            All the n-particle unitary matrices corresponding to the interferometer up
            to cutoff.
        """

    @abc.abstractmethod
    def pfaffian(self, matrix):
        """Calculates the pfaffian of a matrix."""

    @abc.abstractmethod
    def real_logm(self, matrix):
        """Calculates the real logarithm of a matrix."""

    @abc.abstractmethod
    def calculate_interferometer_on_fermionic_fock_space(self, matrix, cutoff):
        """Calculates the representation of the interferometer on the Fock space."""

    @abc.abstractmethod
    def apply_fermionic_passive_linear_to_state_vector(
        self, representations, state_vector, modes, d, cutoff
    ):
        """Applies a passive linear gate to a state vector expressed in the Fock basis."""  # noqa: E501



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import matplotlib.pyplot as plt

import piquasso as pq

np.set_printoptions(suppress=True, linewidth=200)


def main():
    d = 5
    shots = 20000
    mean_photon_number = 1.2

    adjacency_matrix = np.array(
        [
            [0, 1, 1, 0, 1],
            [1, 0, 1, 0, 1],
            [1, 1, 0, 1, 1],
            [0, 0, 1, 0, 1],
            [1, 1, 1, 1, 0],
        ]
    )

    # NOTE: In SF the measurement cutoff is 5, couldn't be changed, and in PQ it
    # corresponds to `measurement_cutoff=6`.
    simulator = pq.GaussianSimulator(d=d, config=pq.Config(measurement_cutoff=6))

    with pq.Program() as program:
        pq.Q() | pq.Graph(adjacency_matrix, mean_photon_number=mean_photon_number)

        pq.Q(0, 1, 2) | pq.ParticleNumberMeasurement()

    result = simulator.execute(program, shots=shots)

    samples = result.samples
    state = result.state
    reduced_state = state.reduced((0, 1, 2))

    samples_as_list = [tuple(x) for x in samples]

    sample_set = set(samples_as_list)

    f_obs = []
    f_exp = []

    for sample in sample_set:
        count = samples_as_list.count(sample)
        if count < 10:
            continue

        f_obs.append(count / shots)
        f_exp.append(reduced_state.get_particle_detection_probability(sample).real)

    f_obs = np.array(f_obs) / np.sum(f_obs)
    f_exp = np.array(f_exp) / np.sum(f_exp)

    plt.bar(np.arange(len(f_obs)), f_exp, fc=(0.5, 0, 0.5, 0.5), label="Expected")
    plt.bar(np.arange(len(f_obs)), f_obs, fc=(0, 1, 0, 0.5), label="Observed")

    plt.xlim(-1, len(f_obs))

    plt.legend()

    plt.get_current_fig_manager().full_screen_toggle()

    plt.show()


if __name__ == "__main__":
    main()



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

import piquasso as pq


def test_str_of_preparations():
    assert str(pq.Vacuum()) == "Vacuum(modes=())"
    assert (
        str(pq.StateVector([1, 2]))
        == "StateVector(occupation_numbers=(1, 2), coefficient=1.0, modes=())"
    )

    assert (
        str(pq.StateVector([1, 2]).on_modes(0, 1))
        == "StateVector(occupation_numbers=(1, 2), coefficient=1.0, modes=(0, 1))"
    )

    assert (
        str(pq.StateVector([1, 2]).on_modes(0, 1) * 0.2)
        == "StateVector(occupation_numbers=(1, 2), coefficient=0.2, modes=(0, 1))"
    )


def test_str_of_gates():
    assert (
        str(pq.Phaseshifter(phi=np.pi / 3))
        == "Phaseshifter(phi=1.0471975511965976, modes=())"
    )

    assert (
        str(pq.Phaseshifter(phi=np.pi / 3).on_modes(1))
        == "Phaseshifter(phi=1.0471975511965976, modes=(1,))"
    )

    assert str(pq.Squeezing(r=1.0)) == "Squeezing(r=1.0, phi=0.0, modes=())"
    assert (
        str(pq.Displacement(r=1.0, phi=0.5)) == "Displacement(r=1.0, phi=0.5, modes=())"
    )

    assert (
        str(pq.Beamsplitter(theta=0.2, phi=0.3))
        == "Beamsplitter(theta=0.2, phi=0.3, modes=())"
    )



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Preparations
============

The built-in preparation instructions in Piquasso. Preparation instructions should
be placed at the beginning of the Piquasso program.
"""


from typing import Dict, Iterable, Tuple, Optional

import numpy as np

from piquasso.core import _mixins
from piquasso.api.instruction import Preparation
from piquasso.api.exceptions import InvalidParameter, InvalidState

from piquasso._math.validations import all_natural, all_real_and_positive


class Vacuum(Preparation):
    r"""Prepare the system in a vacuum state.

    Example usage::

        with pq.Program() as program:
            pq.Q() | pq.Vacuum()
            ...

    Note:
        This operation can only be used for all modes and is only practical to be used
        at the beginning of a program declaration.
    """

    def __init__(self) -> None:
        pass


class Mean(Preparation):
    r"""Set the first canonical moment of the state.

    Example usage::

        with pq.Program() as program:
            pq.Q() | pq.Mean(
                mean=np.array(...)
            )
            ...

    Can only be applied to the following states:
    :class:`~piquasso._simulators.gaussian.state.GaussianState`.

    Note:
        The mean vector is dependent on :math:`\hbar`, but the value of :math:`\hbar`
        is specified later when executed by a simulator. The parameter `mean` should be
        specified keeping in mind that it will automatically be scaled with
        :math:`\hbar` during execution.
    """

    def __init__(self, mean: np.ndarray) -> None:
        """
        Args:
            mean (numpy.ndarray):
                The vector of the first canonical moments in `xpxp`-ordering.
        """
        super().__init__(params=dict(mean=mean))


class Covariance(Preparation):
    r"""Sets the covariance matrix of the state.

    Example usage::

        with pq.Program() as program:
            pq.Q() | pq.Covariance(
                cov=np.array(...)
            )
            ...

    Can only be applied to the following states:
    :class:`~piquasso._simulators.gaussian.state.GaussianState`.

    Note:
        The covariance matrix is dependent on :math:`\hbar`, but the value of
        :math:`\hbar` is specified later when executed by a simulator. The parameter
        `cov` should be specified keeping in mind that it will automatically be scaled
        with :math:`\hbar` during execution.
    """

    def __init__(self, cov: np.ndarray) -> None:
        """
        Args:
            cov (numpy.ndarray): The covariance matrix in `xpxp`-ordering.
        """

        super().__init__(params=dict(cov=cov))


class Thermal(Preparation):
    r"""Prepares a thermal state.

    Example usage::

        with pq.Program() as program:
            pq.Q(0, 1) | pq.Thermal([0.5, 1.5])
            ...

    The thermal state is defined by

    .. math::
        \rho = \sum_{\vec{n} = 0}^\infty
            \frac{\overline{n}^{\vec{n}}}{(1 + \overline{n})^{\vec{n} + 1}}
            | \vec{n} \rangle \langle \vec{n} |,

    where :math:`\overline{n} \in \mathbb{R}^d` is the vector of mean photon numbers.
    The power on vectors is defined as

    .. math::
        \vec{a}^{\vec{b}} = \prod_{i=1}^d a_i^{b_i}.

    In terms of Gaussian states, the mean vector and covariance matrix is

    .. math::
        \mu &= 0_{2d} \\
        \sigma &= \hbar (
            2 \operatorname{diag}(\operatorname{repeat}
                (\overline{n}, 2)) + I_{2d \times 2d}
        ).

    Can only be applied to the following states:
    :class:`~piquasso._simulators.gaussian.state.GaussianState`.
    """

    def __init__(self, mean_photon_numbers: Iterable[float]) -> None:
        """
        Args:
            mean_photon_numbers (Iterable[float]): The sequence of mean photon numbers.

        Raises:
            InvalidParameter: If the mean photon numbers are not positive real numbers.
        """
        if not all_real_and_positive(mean_photon_numbers):
            raise InvalidParameter(
                "The mean photon numbers must be positive real numbers: "
                f"mean_photon_numbers: {mean_photon_numbers}"
            )

        cov = np.diag(2 * np.repeat(np.array(mean_photon_numbers), 2) + 1)

        super().__init__(
            params=dict(mean_photon_numbers=mean_photon_numbers),
            extra_params=dict(cov=cov),
        )


class StateVector(Preparation, _mixins.WeightMixin):
    r"""State preparation with Fock basis vectors.

    Example usage::

        with pq.Program() as program:
            pq.Q() | (
                0.3 * pq.StateVector([2, 1, 0, 3])
                + 0.2 * pq.StateVector([1, 1, 2, 3])
                ...
            )
            ...

    Can only be applied to the following states:
    :class:`~piquasso._simulators.fock.pure.state.PureFockState`.
    """

    def __init__(
        self,
        occupation_numbers: Optional[Iterable[int]] = None,
        fock_amplitude_map: Optional[Dict[Tuple[int, ...], complex]] = None,
        coefficient: complex = 1.0,
    ) -> None:
        """
        Args:
            occupation_numbers (Iterable[int], optional): The occupation numbers.
            fock_amplitude_map (Dict[Tuple[int, ...], complex], optional):
                A mapping of occupation numbers to their corresponding amplitudes.
            coefficient (complex, optional):
                The coefficient of the occupation number. Defaults to :math:`1.0`.

        Raises:
            InvalidParameter:
                If neither `occupation_numbers` nor `fock_amplitude_map` is provided.
            InvalidParameter:
                If both `occupation_numbers` and `fock_amplitude_map` are provided.
            InvalidState:
                If the specified occupation numbers are not all natural numbers.
            InvalidState:
                If the keys in `fock_amplitude_map` are not all natural numbers.
        """

        if occupation_numbers is None and fock_amplitude_map is None:
            raise InvalidParameter(
                "Either 'occupation_numbers' or 'fock_amplitude_map' must be provided."
            )

        if occupation_numbers is not None and fock_amplitude_map is not None:
            raise InvalidParameter(
                "Only one of 'occupation_numbers' or 'fock_amplitude_map' "
                "can be provided."
            )

        if occupation_numbers is not None:
            if not all_natural(occupation_numbers):
                raise InvalidState(
                    f"Invalid occupation numbers: "
                    f"occupation_numbers={occupation_numbers}\n"
                    "Occupation numbers must contain non-negative integers."
                )
            super().__init__(
                params=dict(
                    occupation_numbers=tuple(occupation_numbers),
                    coefficient=coefficient,
                ),
            )

        if fock_amplitude_map is not None:
            if not all(
                all_natural(occupation) for occupation in fock_amplitude_map.keys()
            ):
                raise InvalidState(
                    f"Invalid occupation numbers in "
                    f"fock_amplitude_map: {fock_amplitude_map}\n"
                    "Occupation numbers must contain non-negative integers."
                )
            super().__init__(
                params=dict(
                    fock_amplitude_map=fock_amplitude_map,
                    coefficient=coefficient,
                ),
            )


class DensityMatrix(Preparation, _mixins.WeightMixin):
    r"""State preparation with density matrix elements.

    Example usage::

        with pq.Program() as program:
            pq.Q() | (
                0.2 * pq.DensityMatrix(ket=(1, 0, 1), bra=(2, 1, 0))
                + 0.3 * pq.DensityMatrix(ket=(2, 0, 1), bra=(0, 1, 1))
                ...
            )
            ...

    Note:
        This only creates one matrix element.

    Can only be applied to the following states:
    :class:`~piquasso._simulators.fock.general.state.FockState`.
    """

    def __init__(
        self,
        ket: Iterable[int],
        bra: Iterable[int],
        coefficient: complex = 1.0,
    ) -> None:
        """
        Args:
            bra (Iterable[int]): The bra vector.
            ket (Iterable[int]): The ket vector.
            coefficient (complex, optional):
                The coefficient of the operator defined by the "bra" and "ket" vectors.
                Defaults to :math:`1.0`.

        Raises:
            InvalidState:
                If the specified "bra" or "ket" vectors are not all natural numbers.
        """

        if not all_natural(ket):
            raise InvalidState(f"Invalid ket vector: ket={ket}")

        if not all_natural(bra):
            raise InvalidState(f"Invalid ket vector: ket={bra}")

        super().__init__(
            params=dict(
                ket=tuple(ket),
                bra=tuple(bra),
                coefficient=coefficient,
            ),
        )


class Create(Preparation):
    r"""Create a particle on a mode.

    This instruction essentially applies a creation operator on the specified mode, then
    normalizes the state.

    Example usage::

        with pq.Program() as program:
            pq.Q(1) | pq.Create()
            pq.Q(2) | pq.Create()
            ...

    Can only be applied to the following states:
    :class:`~piquasso._simulators.fock.general.state.FockState`,
    :class:`~piquasso._simulators.fock.pure.state.PureFockState`.
    """

    def __init__(self) -> None:
        pass


class Annihilate(Preparation):
    r"""Annihilate a particle on a mode.

    This instruction essentially applies an annihilation operator on the specified mode,
    then normalizes the state.

    Example usage::

        with pq.Program() as program:
            ...
            pq.Q(0) | pq.Annihilate()
            ...

    Can only be applied to the following states:
    :class:`~piquasso._simulators.fock.general.state.FockState`,
    :class:`~piquasso._simulators.fock.pure.state.PureFockState`.
    """

    def __init__(self) -> None:
        pass



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from piquasso._math.validations import is_natural, all_natural


def test_zero_is_natural():
    assert is_natural(0)


def test_positive_integers_are_natural():
    assert is_natural(2)


def test_negative_integers_are_natural():
    assert not is_natural(-2)


def test_floats_close_to_integers_count_as_natural():
    assert is_natural(2.0)


def test_floats_NOT_close_to_integers_do_NOT_count_as_natural():
    assert not is_natural(2.5)


def test_all_natural_positive_case():
    assert all_natural([1, 1.0, 0.0, 2.0])


def test_all_natural_negative_case():
    assert not all_natural([1, 1.0, 0.0, -2.0])



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import numpy as np

import piquasso as pq


def test_create_number_state():
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(1) | pq.Create()

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 5, phi=np.pi / 6)

    simulator = pq.PureFockSimulator(d=2, config=pq.Config(cutoff=3))

    state = simulator.execute(program).state

    assert np.isclose(state.norm, 1)
    assert np.allclose(
        state.fock_probabilities,
        [0.0, 0.3454915, 0.6545085, 0.0, 0.0, 0.0],
    )


def test_create_and_annihilate_number_state():
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(1) | pq.Create()
        pq.Q(1) | pq.Annihilate()

    simulator = pq.PureFockSimulator(d=2, config=pq.Config(cutoff=3))

    state = simulator.execute(program).state

    assert np.isclose(state.norm, 1)
    assert np.allclose(
        state.fock_probabilities,
        [1, 0, 0, 0, 0, 0],
    )


def test_create_annihilate_and_create():
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

        pq.Q(1) | pq.Create()
        pq.Q(1) | pq.Annihilate()

        pq.Q(1) | pq.Create()

        pq.Q(0, 1) | pq.Beamsplitter(theta=np.pi / 5, phi=np.pi / 6)

    simulator = pq.PureFockSimulator(d=2, config=pq.Config(cutoff=3))

    state = simulator.execute(program).state

    assert np.isclose(state.norm, 1)
    assert np.allclose(
        state.fock_probabilities,
        [0.0, 0.3454915, 0.6545085, 0.0, 0.0, 0.0],
    )


def test_overflow_with_zero_norm_raises_InvalidState_when_normalized():
    with pq.Program() as program:
        pq.Q(2) | pq.StateVector([1]) * np.sqrt(2 / 5)
        pq.Q(1) | pq.StateVector([1]) * np.sqrt(3 / 5)

        pq.Q(1, 2) | pq.Create()

    simulator = pq.PureFockSimulator(d=3, config=pq.Config(cutoff=3))

    state = simulator.execute(program).state

    with pytest.raises(pq.api.exceptions.InvalidState) as error:
        state.normalize()

    assert error.value.args[0] == "The norm of the state is 0."


def test_creation_on_multiple_modes():
    with pq.Program() as program:
        pq.Q(2) | pq.StateVector([1]) * np.sqrt(2 / 5)
        pq.Q(1) | pq.StateVector([1]) * np.sqrt(3 / 5)

        pq.Q(1, 2) | pq.Create()

    simulator = pq.PureFockSimulator(d=3, config=pq.Config(cutoff=4))

    state = simulator.execute(program).state

    assert np.isclose(state.norm, 1)

    assert np.allclose(
        state.fock_probabilities,
        [
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            0.0,
            3 / 5,
            2 / 5,
            0.0,
        ],
    )


def test_state_normalize_after_overflow():
    with pq.Program() as program:
        pq.Q(2) | pq.StateVector([1]) * np.sqrt(2 / 6)
        pq.Q(1) | pq.StateVector([1]) * np.sqrt(3 / 6)
        pq.Q(2) | pq.StateVector([2]) * np.sqrt(1 / 6)

        pq.Q(2) | pq.Create()

    simulator = pq.PureFockSimulator(d=3, config=pq.Config(cutoff=3))

    state = simulator.execute(program).state

    state.normalize()

    assert np.isclose(state.norm, 1)

    assert np.allclose(
        state.fock_probabilities,
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.6, 0.4],
    )


def test_state_vector_with_fock_amplitude_map_preparation():
    amplitude_map = {(0,): 0.6, (1,): 0.8}

    with pq.Program() as program:
        pq.Q() | pq.StateVector(fock_amplitude_map=amplitude_map)

    simulator = pq.PureFockSimulator(d=1, config=pq.Config(cutoff=2))

    state = simulator.execute(program).state

    assert np.allclose(state.state_vector, np.array([0.6, 0.8]))


def test_state_vector_with_fock_amplitude_map_and_coefficient():
    amplitude_map = {(0,): 0.6, (1,): 0.8}
    coefficient = 1 / np.sqrt(2)

    with pq.Program() as program_with_mul:
        pq.Q() | pq.StateVector(fock_amplitude_map=amplitude_map) * coefficient

    with pq.Program() as program_with_param:
        pq.Q() | pq.StateVector(
            fock_amplitude_map=amplitude_map, coefficient=coefficient
        )

    simulator = pq.PureFockSimulator(d=1, config=pq.Config(cutoff=2))

    state_with_mul = simulator.execute(program_with_mul).state
    state_with_param = simulator.execute(program_with_param).state

    expected = coefficient * np.array([0.6, 0.8])

    assert np.allclose(state_with_mul.state_vector, expected)
    assert np.allclose(state_with_param.state_vector, expected)


def test_state_vector_with_fock_amplitude_map_invalid_shape_raises_InvalidState():
    amplitude_map = {(0, 0): 0.6}

    with pq.Program() as program:
        pq.Q() | pq.StateVector(fock_amplitude_map=amplitude_map)

    simulator = pq.PureFockSimulator(d=1, config=pq.Config(cutoff=2, validate=True))
    with pytest.raises(pq.api.exceptions.InvalidState):
        simulator.execute(program)



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

from typing import Callable

from piquasso.api.connector import BaseConnector


def create_single_mode_displacement_gradient(
    r: float,
    phi: float,
    cutoff: int,
    transformation: np.ndarray,
    connector: BaseConnector,
) -> Callable:
    def displacement_matrix_gradient(upstream):
        np = connector.fallback_np

        index_sqrts = np.sqrt(range(cutoff))

        row_sqrts = index_sqrts * np.exp(1j * phi)
        col_sqrts = index_sqrts * np.exp(-1j * phi)

        row_rolled_transformation = np.roll(transformation, 1, axis=0)
        col_rolled_transformation = np.roll(transformation, 1, axis=1)

        row_term = (row_sqrts * row_rolled_transformation.T).T
        col_term = col_sqrts * col_rolled_transformation

        # NOTE: This algorithm rolls the last elements of the transormation matrix to
        # the 0th place, but the 0th element of `row_sqrts` and `col_sqrts` is always
        # zero, so it is fine.
        phi_grad = (row_term + col_term) * r * 1j

        r_grad = -r * transformation + row_term - col_term

        tf = connector._tf
        static_valued = tf.get_static_value(upstream) is not None

        if static_valued:
            upstream = upstream.numpy()
            r_grad_sum = tf.constant(np.real(np.sum(upstream * np.conj(r_grad))))
            phi_grad_sum = tf.constant(np.real(np.sum(upstream * np.conj(phi_grad))))
        else:
            r_grad_sum = tf.math.real(tf.reduce_sum(upstream * tf.math.conj(r_grad)))
            phi_grad_sum = tf.math.real(
                tf.reduce_sum(upstream * tf.math.conj(phi_grad))
            )

        return (r_grad_sum, phi_grad_sum)

    return displacement_matrix_gradient


def create_single_mode_squeezing_gradient(
    r: float,
    phi: float,
    cutoff: int,
    transformation: np.ndarray,
    connector: BaseConnector,
) -> Callable:
    def squeezing_matrix_gradient(upstream):
        np = connector.fallback_np
        tf = connector._tf

        sechr = 1 / np.cosh(r)
        tanhr = np.tanh(r)

        index_sqrts = np.sqrt(np.arange(cutoff, dtype=np.complex128))

        falling_index_sqrts = index_sqrts * np.roll(index_sqrts, 1)

        row_sqrts = falling_index_sqrts * np.exp(1j * phi)
        col_sqrts = falling_index_sqrts * np.exp(-1j * phi)

        row_rolled_transformation = np.roll(transformation, 2, axis=0)
        col_rolled_transformation = np.roll(transformation, 2, axis=1)

        row_term = (row_sqrts * row_rolled_transformation.T).T
        col_term = col_sqrts * col_rolled_transformation

        # NOTE: This algorithm rolls the last and penultimate elements of the
        # transormation matrix to the 1st and 0th place, but the 0th and 1st element of
        # `row_sqrts` and `col_sqrts` is always zero, so it is fine.
        phi_grad = -0.5j * tanhr * (row_term + col_term)

        diagonally_rolled_transformation = np.roll(transformation, (1, 1), axis=(0, 1))

        r_grad = (
            (-tanhr * 0.5) * transformation
            - (sechr * tanhr)
            * np.outer(index_sqrts, index_sqrts)
            * diagonally_rolled_transformation
            - (sechr**2 * 0.5) * (row_term - col_term)
        )

        static_valued = tf.get_static_value(upstream) is not None

        if static_valued:
            upstream = upstream.numpy()
            r_grad_sum = tf.constant(np.real(np.sum(upstream * np.conj(r_grad))))
            phi_grad_sum = tf.constant(np.real(np.sum(upstream * np.conj(phi_grad))))
        else:
            r_grad_sum = tf.math.real(tf.reduce_sum(upstream * tf.math.conj(r_grad)))
            phi_grad_sum = tf.math.real(
                tf.reduce_sum(upstream * tf.math.conj(phi_grad))
            )

        return (r_grad_sum, phi_grad_sum)

    return squeezing_matrix_gradient



#
# Copyright 2021-2025 Budapest Quantum Computing Group
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import numpy as np

import piquasso as pq

from piquasso.api.exceptions import InvalidParameter, InvalidState


def test_Mean_is_automatically_scaled_by_hbar():
    config = pq.Config(hbar=42)

    xpxp_mean_vector = np.array([1, 2])

    with pq.Program() as program:
        pq.Q() | pq.Mean(xpxp_mean_vector)

    simulator = pq.GaussianSimulator(d=1, config=config)

    result = simulator.execute(program)

    assert np.allclose(
        result.state.xpxp_mean_vector, xpxp_mean_vector * np.sqrt(config.hbar)
    )


def test_Covariance_is_automatically_scaled_by_hbar():
    config = pq.Config(hbar=42)

    xpxp_covariance_matrix = np.array(
        [
            [2, 1],
            [1, 2],
        ]
    )

    with pq.Program() as program:
        pq.Q() | pq.Covariance(xpxp_covariance_matrix)

    simulator = pq.GaussianSimulator(d=1, config=config)

    result = simulator.execute(program)

    assert np.allclose(
        result.state.xpxp_covariance_matrix, xpxp_covariance_matrix * config.hbar
    )


def test_Thermal_is_automatically_scaled_by_hbar():
    config = pq.Config(hbar=42)

    mean_photon_numbers = np.array([1, 2])

    with pq.Program() as program:
        pq.Q() | pq.Thermal(mean_photon_numbers)

    simulator = pq.GaussianSimulator(d=2, config=config)

    result = simulator.execute(program)

    assert np.allclose(
        result.state.xpxp_covariance_matrix,
        config.hbar
        * np.array(
            [
                [3, 0, 0, 0],
                [0, 3, 0, 0],
                [0, 0, 5, 0],
                [0, 0, 0, 5],
            ]
        ),
    )


def test_Thermal_with_zero_mean_photon_numbers_yields_Vacuum():
    config = pq.Config(hbar=42)

    mean_photon_numbers = np.array([0, 0])

    with pq.Program() as thermal_program:
        pq.Q() | pq.Thermal(mean_photon_numbers)

    with pq.Program() as vacuum_program:
        pq.Q() | pq.Vacuum()

    simulator = pq.GaussianSimulator(d=2, config=config)

    thermal_state = simulator.execute(thermal_program).state
    vacuum_state = simulator.execute(vacuum_program).state

    assert thermal_state == vacuum_state


def test_state_initialization_with_misshaped_mean():
    misshaped_mean = np.array(
        [
            [1, 2],
            [3, 4],
        ]
    )

    with pq.Program() as program:
        pq.Q() | pq.Mean(misshaped_mean)

    simulator = pq.GaussianSimulator(d=1)

    with pytest.raises(InvalidState):
        simulator.execute(program)


def test_state_initialization_with_misshaped_covariance():
    misshaped_cov = np.array(
        [
            [1, 2, 10000],
            [1, 1, 10000],
        ]
    )

    with pq.Program() as program:
        pq.Q() | pq.Covariance(misshaped_cov)

    simulator = pq.GaussianSimulator(d=1)

    with pytest.raises(InvalidState):
        simulator.execute(program)


def test_state_initialization_with_nonsymmetric_covariance():
    nonsymmetric_cov = np.array(
        [
            [1, 2],
            [1, 1],
        ]
    )

    with pq.Program() as program:
        pq.Q() | pq.Covariance(nonsymmetric_cov)

    simulator = pq.GaussianSimulator(d=1)

    with pytest.raises(InvalidState):
        simulator.execute(program)


def test_state_initialization_with_nonpositive_covariance():
    nonpositive_cov = np.array(
        [
            [1, 0],
            [0, -1],
        ]
    )

    with pq.Program() as program:
        pq.Q() | pq.Covariance(nonpositive_cov)

    simulator = pq.GaussianSimulator(d=1)

    with pytest.raises(InvalidState):
        simulator.execute(program)


def test_Thermal_with_negative_mean_photon_numbers_raises_InvalidParameter():
    mean_photon_numbers = np.array([-1, 1])

    with pytest.raises(InvalidParameter):
        pq.Q() | pq.Thermal(mean_photon_numbers)


def test_vacuum_resets_the_state(state):
    with pq.Program() as program:
        pq.Q() | pq.Vacuum()

    simulator = pq.GaussianSimulator(d=state.d)

    new_state = simulator.execute(program, initial_state=state).state

    assert np.allclose(
        new_state.xpxp_mean_vector,
        np.zeros(2 * new_state.d),
    )
    assert np.allclose(
        new_state.xpxp_covariance_matrix,
        np.identity(2 * new_state.d) * simulator.config.hbar,
    )
